import numpy as np
from numpy.testing import assert_equal

from yt import load
from yt.testing import (
    assert_rel_equal,
    fake_random_ds,
    requires_file,
    requires_module,
)
from yt.units._numpy_wrapper_functions import uconcatenate


def test_ray():
    for nproc in [1, 2, 4, 8]:
        ds = fake_random_ds(64, nprocs=nproc)
        dx = (ds.domain_right_edge - ds.domain_left_edge) / ds.domain_dimensions
        # Three we choose, to get varying vectors, and ten random
        pp1 = np.random.random((3, 13))
        pp2 = np.random.random((3, 13))
        pp1[:, 0] = [0.1, 0.2, 0.3]
        pp2[:, 0] = [0.8, 0.1, 0.4]
        pp1[:, 1] = [0.9, 0.2, 0.3]
        pp2[:, 1] = [0.8, 0.1, 0.4]
        pp1[:, 2] = [0.9, 0.2, 0.9]
        pp2[:, 2] = [0.8, 0.1, 0.4]
        unitary = ds.arr(1.0, "")
        for i in range(pp1.shape[1]):
            p1 = ds.arr(pp1[:, i] + 1e-8 * np.random.random(3), "code_length")
            p2 = ds.arr(pp2[:, i] + 1e-8 * np.random.random(3), "code_length")

            my_ray = ds.ray(p1, p2)
            assert_rel_equal(my_ray["dts"].sum(), unitary, 14)
            ray_cells = my_ray["dts"] > 0

            # find cells intersected by the ray
            my_all = ds.all_data()

            dt = np.abs(dx / (p2 - p1))
            tin = uconcatenate(
                [
                    [(my_all[("index", "x")] - p1[0]) / (p2 - p1)[0] - 0.5 * dt[0]],
                    [(my_all[("index", "y")] - p1[1]) / (p2 - p1)[1] - 0.5 * dt[1]],
                    [(my_all[("index", "z")] - p1[2]) / (p2 - p1)[2] - 0.5 * dt[2]],
                ]
            )
            tout = uconcatenate(
                [
                    [(my_all[("index", "x")] - p1[0]) / (p2 - p1)[0] + 0.5 * dt[0]],
                    [(my_all[("index", "y")] - p1[1]) / (p2 - p1)[1] + 0.5 * dt[1]],
                    [(my_all[("index", "z")] - p1[2]) / (p2 - p1)[2] + 0.5 * dt[2]],
                ]
            )
            tin = tin.max(axis=0)
            tout = tout.min(axis=0)
            my_cells = (tin < tout) & (tin < 1) & (tout > 0)

            assert_equal(ray_cells.sum(), my_cells.sum())
            assert_rel_equal(
                my_ray[("gas", "density")][ray_cells].sum(),
                my_all[("gas", "density")][my_cells].sum(),
                14,
            )
            assert_rel_equal(my_ray["dts"].sum(), unitary, 14)


@requires_module("h5py")
@requires_file("GadgetDiskGalaxy/snapshot_200.hdf5")
def test_ray_particle():
    ds = load("GadgetDiskGalaxy/snapshot_200.hdf5")
    ray = ds.ray(ds.domain_left_edge, ds.domain_right_edge)
    assert_equal(ray["t"].shape, (1451,))
    assert ray["dts"].sum(dtype="f8") > 0



cimport numpy as np
from libc.stdint cimport int32_t, int64_t, uint32_t, uint64_t
from libcpp cimport bool
from libcpp.pair cimport pair
from libcpp.vector cimport vector


cdef extern from "c_utils.hpp":
    double* max_pts(double *pts, uint64_t n, uint64_t m)
    double* min_pts(double *pts, uint64_t n, uint64_t m)
    uint64_t argmax_pts_dim(double *pts, uint64_t *idx,
                            uint32_t m, uint32_t d,
                            uint64_t Lidx, uint64_t Ridx)
    uint64_t argmin_pts_dim(double *pts, uint64_t *idx,
                            uint32_t m, uint32_t d,
                            uint64_t Lidx, uint64_t Ridx)
    void quickSort(double *pts, uint64_t *idx,
                   uint32_t ndim, uint32_t d,
                   int64_t l, int64_t r)
    int64_t partition(double *pts, uint64_t *idx,
                      uint32_t ndim, uint32_t d,
                      int64_t l, int64_t r, int64_t p)
    int64_t partition_given_pivot(double *pts, uint64_t *idx,
                                  uint32_t ndim, uint32_t d,
                                  int64_t l, int64_t r, double pivot)
    int64_t select(double *pts, uint64_t *idx,
                   uint32_t ndim, uint32_t d,
                   int64_t l, int64_t r, int64_t n)
    int64_t pivot(double *pts, uint64_t *idx,
                  uint32_t ndim, uint32_t d,
                  int64_t l, int64_t r)
    void insertSort(double *pts, uint64_t *idx,
                    uint32_t ndim, uint32_t d,
                    int64_t l, int64_t r)
    uint32_t split(double *all_pts, uint64_t *all_idx,
                   uint64_t Lidx, uint64_t n, uint32_t ndim,
                   double *mins, double *maxes,
                   int64_t &split_idx, double &split_val)
    uint32_t split(double *all_pts, uint64_t *all_idx,
                   uint64_t Lidx, uint64_t n, uint32_t ndim,
                   double *mins, double *maxes,
                   int64_t &split_idx, double &split_val,
                   bool use_sliding_midpoint)



import inspect
from collections import Counter
from typing import List, Type

from more_itertools import flatten


def find_lowest_subclasses(candidates: List[Type]) -> List[Type]:
    """
    This function takes a list of classes, and returns only the ones that are
    are not super classes of any others in the list. i.e. the ones that are at
    the bottom of the specified mro of classes.

    Parameters
    ----------
    candidates : Iterable
        An iterable object that is a collection of classes to find the lowest
        subclass of.

    Returns
    -------
    result : list
        A list of classes which are not super classes for any others in
        candidates.
    """
    count = Counter(flatten(inspect.getmro(c) for c in candidates))
    return [x for x in candidates if count[x] == 1]



from yt.testing import assert_allclose_units, requires_file, requires_module
from yt.utilities.answer_testing.framework import data_dir_load

g30 = "IsolatedGalaxy/galaxy0030/galaxy0030"


@requires_module("h5py")
@requires_file(g30)
def test_relative_particle_fields():
    ds = data_dir_load(g30)
    offset = ds.arr([0.1, -0.2, 0.3], "code_length")
    c = ds.domain_center + offset
    sp = ds.sphere(c, (10, "kpc"))
    bv = ds.arr([1.0, 2.0, 3.0], "code_velocity")
    sp.set_field_parameter("bulk_velocity", bv)
    assert_allclose_units(
        sp[("all", "relative_particle_position")], sp[("all", "particle_position")] - c
    )
    assert_allclose_units(
        sp[("all", "relative_particle_velocity")], sp[("all", "particle_velocity")] - bv
    )



import yt

# Load the dataset.
ds = yt.load("Enzo_64/DD0043/data0043")

# Make a density projection centered on the 'm'aximum density location
# with a width of 10 Mpc..
p = yt.ProjectionPlot(ds, "y", ("gas", "density"), center="m", width=(10, "Mpc"))

# Modify the projection
# The argument specifies the region along the line of sight
# for which particles will be gathered.
# 1.0 signifies the entire domain in the line of sight
# p.annotate_particles(1.0)
# but in this case we only go 10 Mpc in depth
p.annotate_particles((10, "Mpc"))

# Save the image.
# Optionally, give a string as an argument
# to name files with a keyword.
p.save()



"""
Title: answer_tests.py
Purpose: Contains answer tests that are used by yt's various frontends
"""
import hashlib
import os
import tempfile

import matplotlib.image as mpimg
import numpy as np

import yt.visualization.plot_window as pw
from yt.utilities.answer_testing.framework import create_obj
from yt.utilities.answer_testing.testing_utilities import (
    _create_phase_plot_attribute_plot,
    _create_plot_window_attribute_plot,
)


def grid_hierarchy(ds):
    result = {}
    result["grid_dimensions"] = ds.index.grid_dimensions
    result["grid_left_edge"] = ds.index.grid_left_edge
    result["grid_right_edge"] = ds.index.grid_right_edge
    result["grid_levels"] = ds.index.grid_levels
    result["grid_particle_count"] = ds.index.grid_particle_count
    return result


def parentage_relationships(ds):
    parents = []
    children = []
    for g in ds.index.grids:
        p = g.Parent
        if p is None:
            parents.append(-1)
        elif hasattr(p, "id"):
            parents.append(p.id)
        else:
            parents = parents + [pg.id for pg in p]
        children = children + [c.id for c in g.Children]
    result = np.array(parents + children)
    return result


def grid_values(ds, field):
    # The hashing is done here so that there is only one entry for
    # the test that contains info about all of the grids as opposed
    # to having a separate 'grid_id : grid_hash' pair for each grid
    # since that makes the answer file much larger
    result = None
    for g in ds.index.grids:
        if result is None:
            result = hashlib.md5(bytes(g.id) + g[field].tobytes())
        else:
            result.update(bytes(g.id) + g[field].tobytes())
        g.clear_data()
    return result.hexdigest()


def projection_values(ds, axis, field, weight_field, dobj_type):
    if dobj_type is not None:
        dobj = create_obj(ds, dobj_type)
    else:
        dobj = None
    if ds.domain_dimensions[axis] == 1:
        # This originally returned None, but None can't be converted
        # to a bytes array (for hashing), so use -1 as a string,
        # since ints can't be converted to bytes either
        return bytes(str(-1).encode("utf-8"))
    proj = ds.proj(field, axis, weight_field=weight_field, data_source=dobj)
    # This is to try and remove python-specific anchors in the yaml
    # answer file. Also, using __repr__() results in weird strings
    # of strings that make comparison fail even though the data is
    # the same
    result = None
    for k, v in proj.field_data.items():
        k = k.__repr__().encode("utf8")
        if result is None:
            result = hashlib.md5(k + v.tobytes())
        else:
            result.update(k + v.tobytes())
    return result.hexdigest()


def field_values(ds, field, obj_type=None, particle_type=False):
    # If needed build an instance of the dataset type
    obj = create_obj(ds, obj_type)
    determined_field = obj._determine_fields(field)[0]
    fd = ds.field_info[determined_field]
    # Get the proper weight field depending on if we're looking at
    # particles or not
    if particle_type:
        weight_field = (determined_field[0], "particle_ones")
    elif fd.is_sph_field:
        weight_field = (determined_field[0], "ones")
    else:
        weight_field = ("index", "ones")
    # Get the average, min, and max
    avg = obj.quantities.weighted_average_quantity(
        determined_field, weight=weight_field
    )
    minimum, maximum = obj.quantities.extrema(field)
    # Return as a hashable bytestring
    return np.array([avg, minimum, maximum])


def pixelized_projection_values(ds, axis, field, weight_field=None, dobj_type=None):
    if dobj_type is not None:
        obj = create_obj(ds, dobj_type)
    else:
        obj = None
    proj = ds.proj(field, axis, weight_field=weight_field, data_source=obj)
    frb = proj.to_frb((1.0, "unitary"), 256)
    frb.render(field)
    if weight_field is not None:
        frb.render(weight_field)
    d = frb.data
    for f in proj.field_data:
        # Sometimes f will be a tuple.
        d[f"{f}_sum"] = proj.field_data[f].sum(dtype="float64")
    # This is to try and remove python-specific anchors in the yaml
    # answer file. Also, using __repr__() results in weird strings
    # of strings that make comparison fail even though the data is
    # the same
    result = None
    for k, v in d.items():
        k = k.__repr__().encode("utf8")
        if result is None:
            result = hashlib.md5(k + v.tobytes())
        else:
            result.update(k + v.tobytes())
    return result.hexdigest()


def small_patch_amr(ds, field, weight, axis, ds_obj):
    hex_digests = {}
    # Grid hierarchy test
    gh_hd = grid_hierarchy(ds)
    hex_digests["grid_hierarchy"] = gh_hd
    # Parentage relationships test
    pr_hd = parentage_relationships(ds)
    hex_digests["parentage_relationships"] = pr_hd
    # Grid values, projection values, and field values tests
    gv_hd = grid_values(ds, field)
    hex_digests["grid_values"] = gv_hd
    fv_hd = field_values(ds, field, ds_obj)
    hex_digests["field_values"] = fv_hd
    pv_hd = projection_values(ds, axis, field, weight, ds_obj)
    hex_digests["projection_values"] = pv_hd
    return hex_digests


def big_patch_amr(ds, field, weight, axis, ds_obj):
    hex_digests = {}
    # Grid hierarchy test
    gh_hd = grid_hierarchy(ds)
    hex_digests["grid_hierarchy"] = gh_hd
    # Parentage relationships test
    pr_hd = parentage_relationships(ds)
    hex_digests["parentage_relationships"] = pr_hd
    # Grid values, projection values, and field values tests
    gv_hd = grid_values(ds, field)
    hex_digests["grid_values"] = gv_hd
    ppv_hd = pixelized_projection_values(ds, axis, field, weight, ds_obj)
    hex_digests["pixelized_projection_values"] = ppv_hd
    return hex_digests


def generic_array(func, args=None, kwargs=None):
    if args is None:
        args = []
    if kwargs is None:
        kwargs = {}
    return func(*args, **kwargs)


def sph_answer(ds, ds_str_repr, ds_nparticles, field, weight, ds_obj, axis):
    # Make sure we're dealing with the right dataset
    assert str(ds) == ds_str_repr
    # Set up keys of test names
    hex_digests = {}
    dd = ds.all_data()
    assert dd["particle_position"].shape == (ds_nparticles, 3)
    tot = sum(
        dd[ptype, "particle_position"].shape[0]
        for ptype in ds.particle_types
        if ptype != "all"
    )
    # Check
    assert tot == ds_nparticles
    dobj = create_obj(ds, ds_obj)
    s1 = dobj["ones"].sum()
    s2 = sum(mask.sum() for block, mask in dobj.blocks)
    assert s1 == s2
    if field[0] in ds.particle_types:
        particle_type = True
    else:
        particle_type = False
    if not particle_type:
        ppv_hd = pixelized_projection_values(ds, axis, field, weight, ds_obj)
        hex_digests["pixelized_projection_values"] = ppv_hd
    fv_hd = field_values(ds, field, ds_obj, particle_type=particle_type)
    hex_digests["field_values"] = fv_hd
    return hex_digests


def get_field_size_and_mean(ds, field, geometric):
    if geometric:
        obj = ds.all_data()
    else:
        obj = ds.data
    return np.array([obj[field].size, obj[field].mean()])


def plot_window_attribute(
    ds,
    plot_field,
    plot_axis,
    attr_name,
    attr_args,
    plot_type="SlicePlot",
    callback_id="",
    callback_runners=None,
):
    if callback_runners is None:
        callback_runners = []
    plot = _create_plot_window_attribute_plot(ds, plot_type, plot_field, plot_axis, {})
    for r in callback_runners:
        r(plot_field, plot)
    attr = getattr(plot, attr_name)
    attr(*attr_args[0], **attr_args[1])
    tmpfd, tmpname = tempfile.mkstemp(suffix=".png")
    os.close(tmpfd)
    plot.save(name=tmpname)
    image = mpimg.imread(tmpname)
    os.remove(tmpname)
    return image


def phase_plot_attribute(
    ds_fn,
    x_field,
    y_field,
    z_field,
    attr_name,
    attr_args,
    plot_type="PhasePlot",
    plot_kwargs=None,
):
    if plot_kwargs is None:
        plot_kwargs = {}
    data_source = ds_fn.all_data()
    plot = _create_phase_plot_attribute_plot(
        data_source, x_field, y_field, z_field, plot_type, plot_kwargs
    )
    attr = getattr(plot, attr_name)
    attr(*attr_args[0], **attr_args[1])
    tmpfd, tmpname = tempfile.mkstemp(suffix=".png")
    os.close(tmpfd)
    plot.save(name=tmpname)
    image = mpimg.imread(tmpname)
    os.remove(tmpname)
    return image


def generic_image(img_fname):
    img_data = mpimg.imread(img_fname)
    return img_data


def axial_pixelization(ds):
    r"""
    This test is typically used once per geometry or coordinates type.
    Feed it a dataset, and it checks that the results of basic pixelization
    don't change.
    """
    for i, axis in enumerate(ds.coordinates.axis_order):
        (bounds, center, display_center) = pw.get_window_parameters(
            axis, ds.domain_center, None, ds
        )
        slc = ds.slice(axis, center[i])
        xax = ds.coordinates.axis_name[ds.coordinates.x_axis[axis]]
        yax = ds.coordinates.axis_name[ds.coordinates.y_axis[axis]]
        pix_x = ds.coordinates.pixelize(axis, slc, xax, bounds, (512, 512))
        pix_y = ds.coordinates.pixelize(axis, slc, yax, bounds, (512, 512))
        # Wipe out all NaNs
        pix_x[np.isnan(pix_x)] = 0.0
        pix_y[np.isnan(pix_y)] = 0.0
        pix_x
        pix_y
    return pix_x, pix_y


def extract_connected_sets(ds_fn, data_source, field, num_levels, min_val, max_val):
    n, all_sets = data_source.extract_connected_sets(
        field, num_levels, min_val, max_val
    )
    result = []
    for level in all_sets:
        for set_id in all_sets[level]:
            result.append(
                [
                    all_sets[level][set_id]["cell_mass"].size,
                    all_sets[level][set_id]["cell_mass"].sum(),
                ]
            )
    result = np.array(result)
    return result


def VR_image_comparison(scene):
    tmpfd, tmpname = tempfile.mkstemp(suffix=".png")
    os.close(tmpfd)
    scene.save(tmpname, sigma_clip=1.0)
    image = mpimg.imread(tmpname)
    os.remove(tmpname)
    return image



import numpy as np
from numpy.testing import assert_array_equal

from yt.testing import fake_amr_ds, fake_random_ds
from yt.units._numpy_wrapper_functions import uintersect1d
from yt.units.yt_array import YTArray


def setup():
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


# Copied from test_boolean for computing a unique identifier for
# each cell from cell positions
def _IDFIELD(field, data):
    width = data.ds.domain_right_edge - data.ds.domain_left_edge
    min_dx = YTArray(1.0 / 8192, units="code_length", registry=data.ds.unit_registry)
    delta = width / min_dx
    x = data[("index", "x")] - min_dx / 2.0
    y = data[("index", "y")] - min_dx / 2.0
    z = data[("index", "z")] - min_dx / 2.0
    xi = x / min_dx
    yi = y / min_dx
    zi = z / min_dx
    index = xi + delta[0] * (yi + delta[1] * zi)
    return index


def test_compose_no_overlap():
    r"""Test to make sure that composed data objects that don't
    overlap behave the way we expect (return empty collections)
    """
    empty = np.array([])
    for n in [1, 2, 4, 8]:
        ds = fake_random_ds(64, nprocs=n)
        ds.add_field(("index", "ID"), sampling_type="cell", function=_IDFIELD)

        # position parameters for initial region
        center = [0.25] * 3
        left_edge = [0.1] * 3
        right_edge = [0.4] * 3
        normal = [1, 0, 0]
        radius = height = 0.15

        # initial 3D regions
        sources = [
            ds.sphere(center, radius),
            ds.region(center, left_edge, right_edge),
            ds.disk(center, normal, radius, height),
        ]

        # position parameters for non-overlapping regions
        center = [0.75] * 3
        left_edge = [0.6] * 3
        right_edge = [0.9] * 3

        # subselect non-overlapping 0, 1, 2, 3D regions
        for data1 in sources:
            data2 = ds.sphere(center, radius, data_source=data1)
            assert_array_equal(data2["index", "ID"], empty)

            data2 = ds.region(center, left_edge, right_edge, data_source=data1)
            assert_array_equal(data2["index", "ID"], empty)

            data2 = ds.disk(center, normal, radius, height, data_source=data1)
            assert_array_equal(data2["index", "ID"], empty)

            for d in range(3):
                data2 = ds.slice(d, center[d], data_source=data1)
                assert_array_equal(data2["index", "ID"], empty)

            for d in range(3):
                data2 = ds.ortho_ray(
                    d, center[0:d] + center[d + 1 :], data_source=data1
                )
                assert_array_equal(data2["index", "ID"], empty)

            data2 = ds.point(center, data_source=data1)
            assert_array_equal(data2["index", "ID"], empty)


def test_compose_overlap():
    r"""Test to make sure that composed data objects that do
    overlap behave the way we expect
    """
    for n in [1, 2, 4, 8]:
        ds = fake_random_ds(64, nprocs=n)
        ds.add_field(("index", "ID"), sampling_type="cell", function=_IDFIELD)

        # position parameters for initial region
        center = [0.4, 0.5, 0.5]
        left_edge = [0.1] * 3
        right_edge = [0.7] * 3
        normal = [1, 0, 0]
        radius = height = 0.15

        # initial 3D regions
        sources = [
            ds.sphere(center, radius),
            ds.region(center, left_edge, right_edge),
            ds.disk(center, normal, radius, height),
        ]

        # position parameters for overlapping regions
        center = [0.6, 0.5, 0.5]
        left_edge = [0.3] * 3
        right_edge = [0.9] * 3

        # subselect non-overlapping 0, 1, 2, 3D regions
        for data1 in sources:
            id1 = data1["index", "ID"]

            data2 = ds.sphere(center, radius)
            data3 = ds.sphere(center, radius, data_source=data1)
            id2 = data2["index", "ID"]
            id3 = data3["index", "ID"]
            id3.sort()
            assert_array_equal(uintersect1d(id1, id2), id3)

            data2 = ds.region(center, left_edge, right_edge)
            data3 = ds.region(center, left_edge, right_edge, data_source=data1)
            id2 = data2["index", "ID"]
            id3 = data3["index", "ID"]
            id3.sort()
            assert_array_equal(uintersect1d(id1, id2), id3)

            data2 = ds.disk(center, normal, radius, height)
            data3 = ds.disk(center, normal, radius, height, data_source=data1)
            id2 = data2["index", "ID"]
            id3 = data3["index", "ID"]
            id3.sort()
            assert_array_equal(uintersect1d(id1, id2), id3)

            for d in range(3):
                data2 = ds.slice(d, center[d])
                data3 = ds.slice(d, center[d], data_source=data1)
                id2 = data2["index", "ID"]
                id3 = data3["index", "ID"]
                id3.sort()
                assert_array_equal(uintersect1d(id1, id2), id3)

            for d in range(3):
                data2 = ds.ortho_ray(d, center[0:d] + center[d + 1 :])
                data3 = ds.ortho_ray(
                    d, center[0:d] + center[d + 1 :], data_source=data1
                )
                id2 = data2["index", "ID"]
                id3 = data3["index", "ID"]
                id3.sort()
                assert_array_equal(uintersect1d(id1, id2), id3)

            data2 = ds.point(center)
            data3 = ds.point(center, data_source=data1)
            id2 = data2["index", "ID"]
            id3 = data3["index", "ID"]
            id3.sort()
            assert_array_equal(uintersect1d(id1, id2), id3)


def test_compose_max_level_min_level():
    ds = fake_amr_ds()
    ad = ds.all_data()
    ad.max_level = 2
    slc = ds.slice("x", 0.5, data_source=ad)
    assert slc[("index", "grid_level")].max() == 2
    frb = slc.to_frb(1.0, 128)
    assert np.all(frb[("stream", "Density")] > 0)
    assert frb[("index", "grid_level")].max() == 2



import sys
from functools import wraps
from importlib.util import find_spec
from typing import Optional, Type


class NotAModule:
    """
    A class to implement an informative error message that will be outputted if
    someone tries to use an on-demand import without having the requisite
    package installed.
    """

    def __init__(self, pkg_name, exc: Optional[BaseException] = None):
        self.pkg_name = pkg_name
        self._original_exception = exc
        error_note = (
            f"Something went wrong while trying to lazy-import {pkg_name}. "
            f"Please make sure that {pkg_name} is properly installed.\n"
            "If the problem persists, please file an issue at "
            "https://github.com/yt-project/yt/issues/new"
        )
        self.error: BaseException
        if exc is None:
            self.error = ImportError(error_note)
        elif sys.version_info >= (3, 11):
            exc.add_note(error_note)
            self.error = exc
        else:
            # mimic Python 3.11 behaviour:
            # preserve error message and traceback
            self.error = type(exc)(f"{exc!s}\n{error_note}").with_traceback(
                exc.__traceback__
            )

    def __getattr__(self, item):
        raise self.error

    def __call__(self, *args, **kwargs):
        raise self.error

    def __repr__(self) -> str:
        if self._original_exception is None:
            return f"NotAModule({self.pkg_name!r})"
        else:
            return f"NotAModule({self.pkg_name!r}, {self._original_exception!r})"


class OnDemand:
    _default_factory: Type[NotAModule] = NotAModule

    def __init_subclass__(cls):
        if not cls.__name__.endswith("_imports"):
            raise TypeError(f"class {cls}'s name needs to be suffixed '_imports'")

    def __new__(cls):
        if cls is OnDemand:
            raise TypeError("The OnDemand base class cannot be instantiated.")
        else:
            return object.__new__(cls)

    @property
    def _name(self) -> str:
        _name, _, _suffix = self.__class__.__name__.rpartition("_")
        return _name

    @property
    def __is_available__(self) -> bool:
        # special protocol to support testing framework
        return find_spec(self._name) is not None


def safe_import(func):
    @property
    @wraps(func)
    def inner(self):
        try:
            return func(self)
        except ImportError as exc:
            return self._default_factory(self._name, exc)

    return inner


class netCDF4_imports(OnDemand):
    def __init__(self):
        # this ensures the import ordering between netcdf4 and h5py. If h5py is
        # imported first, can get file lock errors on some systems (including travis-ci)
        # so we need to do this before initializing h5py_imports()!
        # similar to this issue https://github.com/pydata/xarray/issues/2560
        try:
            import netCDF4  # noqa F401
        except ImportError:
            pass

    @safe_import
    def Dataset(self):
        from netCDF4 import Dataset

        return Dataset


_netCDF4 = netCDF4_imports()


class astropy_imports(OnDemand):
    @safe_import
    def log(self):
        from astropy import log

        if log.exception_logging_enabled():
            log.disable_exception_logging()

        return log

    @safe_import
    def pyfits(self):
        from astropy.io import fits

        return fits

    @safe_import
    def pywcs(self):
        import astropy.wcs as pywcs

        self.log
        return pywcs

    @safe_import
    def units(self):
        from astropy import units

        self.log
        return units

    @safe_import
    def conv(self):
        import astropy.convolution as conv

        self.log
        return conv

    @safe_import
    def time(self):
        import astropy.time as time

        self.log
        return time

    @safe_import
    def wcsaxes(self):
        from astropy.visualization import wcsaxes

        self.log
        return wcsaxes

    @safe_import
    def WCS(self):
        from astropy.wcs import WCS

        self.log
        return WCS


_astropy = astropy_imports()


class regions_imports(OnDemand):
    @safe_import
    def Regions(self):
        from regions import Regions

        return Regions


_regions = regions_imports()


class NotCartopy(NotAModule):
    """
    A custom class to return error messages dependent on system installation
    for cartopy imports.
    """

    def __init__(self, pkg_name):
        self.pkg_name = pkg_name
        if any(s in sys.version for s in ("Anaconda", "Continuum")):
            # the conda-based installs of cartopy don't have issues with the
            # GEOS library, so the error message for users with conda can be
            # relatively short. Discussion related to this is in
            # yt-project/yt#1966
            self.error = ImportError(
                "This functionality requires the %s "
                "package to be installed." % self.pkg_name
            )
        else:
            self.error = ImportError(
                f"This functionality requires the {pkg_name} "
                "package to be installed.\n"
                "For further instruction please refer to Cartopy's documentation\n"
                "https://scitools.org.uk/cartopy/docs/latest/installing.html"
            )


class cartopy_imports(OnDemand):
    _default_factory = NotCartopy

    @safe_import
    def crs(self):
        import cartopy.crs as crs

        return crs


_cartopy = cartopy_imports()


class pooch_imports(OnDemand):
    @safe_import
    def HTTPDownloader(self):
        from pooch import HTTPDownloader

        return HTTPDownloader

    @safe_import
    def utils(self):
        from pooch import utils

        return utils

    @safe_import
    def create(self):
        from pooch import create

        return create


_pooch = pooch_imports()


class pyart_imports(OnDemand):
    @safe_import
    def io(self):
        from pyart import io

        return io

    @safe_import
    def map(self):
        from pyart import map

        return map


_pyart = pyart_imports()


class xarray_imports(OnDemand):
    @safe_import
    def open_dataset(self):
        from xarray import open_dataset

        return open_dataset


_xarray = xarray_imports()


class scipy_imports(OnDemand):
    @safe_import
    def signal(self):
        from scipy import signal

        return signal

    @safe_import
    def spatial(self):
        from scipy import spatial

        return spatial

    @safe_import
    def ndimage(self):
        from scipy import ndimage

        return ndimage

    # Optimize is not presently used by yt, but appears here to enable
    # required functionality in yt extension, trident

    @safe_import
    def optimize(self):
        from scipy import optimize

        return optimize


_scipy = scipy_imports()


class h5py_imports(OnDemand):
    @safe_import
    def File(self):
        from h5py import File

        return File

    @safe_import
    def Group(self):
        from h5py import Group

        return Group

    @safe_import
    def Dataset(self):
        from h5py import Dataset

        return Dataset

    @safe_import
    def get_config(self):
        from h5py import get_config

        return get_config

    @safe_import
    def h5f(self):
        from h5py import h5f

        return h5f

    @safe_import
    def h5p(self):
        from h5py import h5p

        return h5p

    @safe_import
    def h5d(self):
        from h5py import h5d

        return h5d

    @safe_import
    def h5s(self):
        from h5py import h5s

        return h5s


_h5py = h5py_imports()


class nose_imports(OnDemand):
    @safe_import
    def run(self):
        from nose import run

        return run


_nose = nose_imports()


class libconf_imports(OnDemand):
    @safe_import
    def load(self):
        from libconf import load

        return load


_libconf = libconf_imports()


class yaml_imports(OnDemand):
    @safe_import
    def load(self):
        from yaml import load

        return load

    @safe_import
    def FullLoader(self):
        from yaml import FullLoader

        return FullLoader


_yaml = yaml_imports()


class NotMiniball(NotAModule):
    def __init__(self, pkg_name):
        super().__init__(pkg_name)
        str = (
            "This functionality requires the %s package to be installed. "
            "Installation instructions can be found at "
            "https://github.com/weddige/miniball or alternatively you can "
            "install via `python -m pip install MiniballCpp`."
        )
        self.error = ImportError(str % self.pkg_name)


class miniball_imports(OnDemand):
    @safe_import
    def Miniball(self):
        from miniball import Miniball

        return Miniball


_miniball = miniball_imports()


class f90nml_imports(OnDemand):
    @safe_import
    def read(self):
        from f90nml import read

        return read

    @safe_import
    def Namelist(self):
        from f90nml import Namelist

        return Namelist


_f90nml = f90nml_imports()


class requests_imports(OnDemand):
    @safe_import
    def post(self):
        from requests import post

        return post

    @safe_import
    def put(self):
        from requests import put

        return put

    @safe_import
    def codes(self):
        from requests import codes

        return codes

    @safe_import
    def get(self):
        from requests import get

        return get

    @safe_import
    def exceptions(self):
        from requests import exceptions

        return exceptions


_requests = requests_imports()


class pandas_imports(OnDemand):
    @safe_import
    def NA(self):
        from pandas import NA

        return NA

    @safe_import
    def DataFrame(self):
        from pandas import DataFrame

        return DataFrame

    @safe_import
    def concat(self):
        from pandas import concat

        return concat


_pandas = pandas_imports()


class firefly_imports(OnDemand):
    @safe_import
    def data_reader(self):
        import firefly.data_reader as data_reader

        return data_reader

    @safe_import
    def server(self):
        import firefly.server as server

        return server


_firefly = firefly_imports()


# Note: ratarmount may fail with an OSError on import if libfuse is missing
class ratarmount_imports(OnDemand):
    @safe_import
    def TarMount(self):
        from ratarmount import TarMount

        return TarMount

    @safe_import
    def fuse(self):
        from ratarmount import fuse

        return fuse


_ratarmount = ratarmount_imports()



from abc import ABC, abstractmethod

from more_itertools import always_iterable


class Union(ABC):
    @property
    @abstractmethod
    def _union_type(self) -> str:
        ...

    def __init__(self, name, sub_types):
        self.name = name
        self.sub_types = list(always_iterable(sub_types))

    def __iter__(self):
        yield from self.sub_types

    def __repr__(self):
        return "{} Union: '{}' composed of: {}".format(
            self._union_type.capitalize(), self.name, self.sub_types
        )


class MeshUnion(Union):
    _union_type = "mesh"


class ParticleUnion(Union):
    _union_type = "particle"



import numpy as np
from numpy.testing import assert_array_equal, assert_equal

from yt.frontends.gadget_fof.api import GadgetFOFDataset
from yt.testing import ParticleSelectionComparison, requires_file, requires_module
from yt.utilities.answer_testing.framework import (
    FieldValuesTest,
    data_dir_load,
    requires_ds,
)

_fields = (
    ("all", "particle_position_x"),
    ("all", "particle_position_y"),
    ("all", "particle_position_z"),
    ("all", "particle_velocity_x"),
    ("all", "particle_velocity_y"),
    ("all", "particle_velocity_z"),
    ("all", "particle_mass"),
    ("all", "particle_identifier"),
)

# a dataset with empty files
g5 = "gadget_fof_halos/groups_005/fof_subhalo_tab_005.0.hdf5"
g42 = "gadget_fof_halos/groups_042/fof_subhalo_tab_042.0.hdf5"


@requires_module("h5py")
@requires_ds(g5)
def test_fields_g5():
    for field in _fields:
        yield FieldValuesTest(g5, field, particle_type=True)


@requires_module("h5py")
@requires_ds(g42)
def test_fields_g42():
    for field in _fields:
        yield FieldValuesTest(g42, field, particle_type=True)


@requires_module("h5py")
@requires_file(g42)
def test_GadgetFOFDataset():
    assert isinstance(data_dir_load(g42), GadgetFOFDataset)


# fof/subhalo catalog with member particles
g298 = "gadget_halos/data/groups_298/fof_subhalo_tab_298.0.hdf5"


@requires_module("h5py")
@requires_file(g298)
def test_particle_selection():
    ds = data_dir_load(g298)
    psc = ParticleSelectionComparison(ds)
    psc.run_defaults()


@requires_module("h5py")
@requires_file(g298)
def test_subhalos():
    ds = data_dir_load(g298)
    total_sub = 0
    total_int = 0
    for hid in range(0, ds.index.particle_count["Group"]):
        my_h = ds.halo("Group", hid)
        h_ids = my_h[("Group", "ID")]
        for sid in range(int(my_h[("Group", "subhalo_number")][0])):
            my_s = ds.halo("Subhalo", (my_h.particle_identifier, sid))
            total_sub += my_s[("Subhalo", "ID")].size
            total_int += np.intersect1d(h_ids, my_s[("Subhalo", "ID")]).size

    # Test that all subhalo particles are contained within
    # their parent group.
    assert_equal(total_sub, total_int)


@requires_module("h5py")
@requires_file(g298)
def test_halo_masses():
    ds = data_dir_load(g298)
    ad = ds.all_data()
    for ptype in ["Group", "Subhalo"]:
        nhalos = ds.index.particle_count[ptype]
        mass = ds.arr(np.zeros(nhalos), "code_mass")
        for i in range(nhalos):
            halo = ds.halo(ptype, i)
            mass[i] = halo.mass

        # Check that masses from halo containers are the same
        # as the array of all masses.  This will test getting
        # scalar fields for halos correctly.
        assert_array_equal(ad[ptype, "particle_mass"], mass)


# fof/subhalo catalog with no member ids in first file
g56 = "gadget_halos/data/groups_056/fof_subhalo_tab_056.0.hdf5"


# This dataset has halos in one file and ids in another,
# which can confuse the field detection.
@requires_module("h5py")
@requires_file(g56)
def test_unbalanced_dataset():
    ds = data_dir_load(g56)
    halo = ds.halo("Group", 0)
    assert_equal(len(halo[("Group", "member_ids")]), 33)
    assert_equal(halo[("Group", "member_ids")].min().d, 723254.0)
    assert_equal(halo[("Group", "member_ids")].max().d, 772662.0)


# fof/subhalo catalog with no member ids in first file
g76 = "gadget_halos/data/groups_076/fof_subhalo_tab_076.0.hdf5"


# This dataset has one halo with particles distributed over 3 files
# with the 2nd file being empty.
@requires_module("h5py")
@requires_file(g76)
def test_3file_halo():
    ds = data_dir_load(g76)
    # this halo's particles are distributed over 3 files with the
    # middle file being empty
    halo = ds.halo("Group", 6)
    halo[("Group", "member_ids")]
    assert True



"""
CF Radial data structures



"""
import contextlib
import os
import weakref
from typing import Optional, Tuple

import numpy as np
from unyt import unyt_array

from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.static_output import Dataset
from yt.funcs import mylog
from yt.geometry.grid_geometry_handler import GridIndex
from yt.utilities.file_handler import NetCDF4FileHandler, warn_netcdf
from yt.utilities.on_demand_imports import _xarray as xr

from .fields import CFRadialFieldInfo


class CFRadialGrid(AMRGridPatch):
    _id_offset = 0

    def __init__(self, id, index, level, dimensions):
        super().__init__(id, filename=index.index_filename, index=index)
        self.Parent = None
        self.Children = []
        self.Level = level
        self.ActiveDimensions = dimensions


class CFRadialHierarchy(GridIndex):
    grid = CFRadialGrid

    def __init__(self, ds, dataset_type="cf_radial"):
        self.dataset_type = dataset_type
        self.dataset = weakref.proxy(ds)
        # our index file is the dataset itself:
        self.index_filename = self.dataset.parameter_filename
        self.directory = os.path.dirname(self.index_filename)
        # float type for the simulation edges and must be float64 now
        self.float_type = np.float64
        super().__init__(ds, dataset_type)

    def _detect_output_fields(self):
        # This sets self.field_list, containing all the available on-disk fields and
        # records the units for each field.
        self.field_list = []
        units = {}
        with self.ds._handle() as xr_ds_handle:
            for key in xr_ds_handle.variables.keys():
                if all(x in xr_ds_handle[key].dims for x in ["time", "z", "y", "x"]):
                    fld = ("cf_radial", key)
                    self.field_list.append(fld)
                    units[fld] = xr_ds_handle[key].units

        self.ds.field_units.update(units)

    def _count_grids(self):
        self.num_grids = 1

    def _parse_index(self):
        self.grid_left_edge[0][:] = self.ds.domain_left_edge[:]
        self.grid_right_edge[0][:] = self.ds.domain_right_edge[:]
        self.grid_dimensions[0][:] = self.ds.domain_dimensions[:]
        self.grid_particle_count[0][0] = 0
        self.grid_levels[0][0] = 0
        self.max_level = 0

    def _populate_grid_objects(self):
        # only a single grid, no need to loop
        g = self.grid(0, self, self.grid_levels.flat[0], self.grid_dimensions[0])
        g._prepare_grid()
        g._setup_dx()
        self.grids = np.array([g], dtype="object")


class CFRadialDataset(Dataset):
    _index_class = CFRadialHierarchy
    _field_info_class = CFRadialFieldInfo

    def __init__(
        self,
        filename,
        dataset_type="cf_radial",
        storage_filename=None,
        storage_overwrite: bool = False,
        grid_shape: Optional[Tuple[int, int, int]] = None,
        grid_limit_x: Optional[Tuple[float, float]] = None,
        grid_limit_y: Optional[Tuple[float, float]] = None,
        grid_limit_z: Optional[Tuple[float, float]] = None,
        units_override=None,
    ):
        """

        Parameters
        ----------
        filename
        dataset_type
        storage_filename: Optional[str]
            the filename to store gridded file to if necessary. If not provided,
            the string "_yt_grid" will be appended to the dataset filename.
        storage_overwrite: bool
            if True and if any gridding parameters are set, then the
            storage_filename will be over-written if it exists. Default is False.
        grid_shape : Optional[Tuple[int, int, int]]
            when gridding to cartesian, grid_shape is the number of cells in the
            z, y, x coordinates. If not provided, yt attempts to calculate a
            reasonable shape based on the resolution of the original cfradial grid
        grid_limit_x : Optional[Tuple[float, float]]
            The x range of the cartesian-gridded data in the form (xmin, xmax) with
            x in the native radar range units
        grid_limit_y : Optional[Tuple[float, float]]
            The y range of the cartesian-gridded data in the form (ymin, ymax) with
            y in the native radar range units
        grid_limit_z : Optional[Tuple[float, float]]
            The z range of the cartesian-gridded data in the form (zmin, zmax) with
            z in the native radar range units
        units_override
        """
        self.fluid_types += ("cf_radial",)

        with self._handle(filename=filename) as xr_ds_handle:
            if "x" not in xr_ds_handle.coords:
                if storage_filename is None:
                    f_base, f_ext = os.path.splitext(filename)
                    storage_filename = f_base + "_yt_grid" + f_ext

                regrid = True
                if os.path.exists(storage_filename):
                    # pyart grid.write will error if the filename exists, so this logic
                    # forces some explicit behavior to minimize confusion and avoid
                    # overwriting or deleting without explicit user consent.
                    if storage_overwrite:
                        os.remove(storage_filename)
                    elif any([grid_shape, grid_limit_x, grid_limit_y, grid_limit_z]):
                        mylog.warning(
                            "Ignoring provided grid parameters because %s exists.",
                            storage_filename,
                        )
                        mylog.warning(
                            "To re-grid, either provide a unique storage_filename or set "
                            "storage_overwrite to True to overwrite %s.",
                            storage_filename,
                        )
                        regrid = False
                    else:
                        mylog.info(
                            "loading existing re-gridded file: %s", storage_filename
                        )
                        regrid = False

                if regrid:
                    mylog.info("Building cfradial grid")
                    from yt.utilities.on_demand_imports import _pyart as pyart

                    radar = pyart.io.read_cfradial(filename)

                    grid_limit_z = self._validate_grid_dim(radar, "z", grid_limit_z)
                    grid_limit_x = self._validate_grid_dim(radar, "x", grid_limit_x)
                    grid_limit_y = self._validate_grid_dim(radar, "y", grid_limit_y)
                    grid_limits = (grid_limit_z, grid_limit_y, grid_limit_x)
                    grid_shape = self._validate_grid_shape(grid_shape)
                    # note: grid_shape must be in (z, y, x) order.

                    self.grid_shape = grid_shape
                    self.grid_limits = grid_limits
                    mylog.info("Calling pyart.map.grid_from_radars ... ")
                    # this is fairly slow
                    grid = pyart.map.grid_from_radars(
                        (radar,),
                        grid_shape=self.grid_shape,
                        grid_limits=self.grid_limits,
                    )
                    mylog.info(
                        "Successfully built cfradial grid, writing to %s",
                        storage_filename,
                    )
                    mylog.info(
                        "Subsequent loads of %s will load the gridded file by default",
                        filename,
                    )
                    grid.write(storage_filename)

                filename = storage_filename

        super().__init__(filename, dataset_type, units_override=units_override)
        self.storage_filename = storage_filename
        self.refine_by = 2  # refinement factor between a grid and its subgrid

    @contextlib.contextmanager
    def _handle(self, filename: Optional[str] = None):
        if filename is None:
            if hasattr(self, "filename"):
                filename = self.filename
            else:
                raise RuntimeError("Dataset has no filename yet.")

        with xr.open_dataset(filename) as xrds:
            yield xrds

    def _validate_grid_dim(
        self, radar, dim: str, grid_limit: Optional[Tuple[float, float]] = None
    ) -> Tuple[float, float]:
        if grid_limit is None:
            if dim.lower() == "z":
                gate_alt = radar.gate_altitude["data"]
                gate_alt_units = radar.gate_altitude["units"]
                grid_limit = (gate_alt.min(), gate_alt.max())
                grid_limit = self._round_grid_guess(grid_limit, gate_alt_units)
                mylog.info(
                    "grid_limit_z not provided, using max height range in data: (%f, %f)",
                    *grid_limit,
                )
            else:
                max_range = radar.range["data"].max()
                grid_limit = self._round_grid_guess(
                    (-max_range, max_range), radar.range["units"]
                )
                mylog.info(
                    "grid_limit_%s not provided, using max horizontal range in data: (%f, %f)",
                    dim,
                    *grid_limit,
                )
        if len(grid_limit) != 2:
            raise ValueError(
                f"grid_limit_{dim} must have 2 dimensions, but it has {len(grid_limit)}"
            )

        return grid_limit

    def _validate_grid_shape(
        self, grid_shape: Optional[Tuple[int, int, int]] = None
    ) -> Tuple[int, int, int]:
        if grid_shape is None:
            grid_shape = (100, 100, 100)
            mylog.info(
                "grid_shape not provided, using (nz, ny, nx) = (%i, %i, %i)",
                *grid_shape,
            )
        if len(grid_shape) != 3:
            raise ValueError(
                f"grid_shape must have 3 dimensions, but it has {len(grid_shape)}"
            )
        return grid_shape

    def _round_grid_guess(self, bounds: Tuple[float, float], unit_str: str):
        # rounds the bounds to the closest 10 km increment that still contains
        # the grid_limit
        for findstr, repstr in self._field_info_class.unit_subs:
            unit_str = unit_str.replace(findstr, repstr)
        limits = unyt_array(bounds, unit_str).to("km")
        limits[0] = np.floor(limits[0] / 10.0) * 10.0
        limits[1] = np.ceil(limits[1] / 10.0) * 10.0
        return tuple(limits.to(unit_str).tolist())

    def _set_code_unit_attributes(self):
        with self._handle() as xr_ds_handle:
            length_unit = xr_ds_handle.variables["x"].attrs["units"]
        self.length_unit = self.quan(1.0, length_unit)
        self.mass_unit = self.quan(1.0, "kg")
        self.time_unit = self.quan(1.0, "s")

    def _parse_parameter_file(self):
        self.parameters = {}

        with self._handle() as xr_ds_handle:
            x, y, z = (xr_ds_handle.coords[d] for d in "xyz")

            self.origin_latitude = xr_ds_handle.origin_latitude[0]
            self.origin_longitude = xr_ds_handle.origin_longitude[0]

            self.domain_left_edge = np.array([x.min(), y.min(), z.min()])
            self.domain_right_edge = np.array([x.max(), y.max(), z.max()])
            self.dimensionality = 3
            dims = [xr_ds_handle.dims[d] for d in "xyz"]
            self.domain_dimensions = np.array(dims, dtype="int64")
            self._periodicity = (False, False, False)
            self.current_time = float(xr_ds_handle.time.values)

        # Cosmological information set to zero (not in space).
        self.cosmological_simulation = 0
        self.current_redshift = 0.0
        self.omega_lambda = 0.0
        self.omega_matter = 0.0
        self.hubble_constant = 0.0

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        # This accepts a filename or a set of arguments and returns True or
        # False depending on if the file is of the type requested.

        if not xr.__is_available__:
            return False

        warn_netcdf(filename)
        is_cfrad = False
        try:
            # note that we use the NetCDF4FileHandler here to avoid some
            # issues with xarray opening datasets it cannot handle. Once
            # a dataset is as identified as a CFRadialDataset, xarray is used
            # for opening. See https://github.com/yt-project/yt/issues/3987
            nc4_file = NetCDF4FileHandler(filename)
            with nc4_file.open_ds(keepweakref=True) as ds:
                con = "Conventions"  # the attribute to check for file conventions
                cons = ""  # the value of the Conventions attribute
                for c in [con, con.lower()]:
                    if hasattr(ds, c):
                        cons += getattr(ds, c)
                is_cfrad = "CF/Radial" in cons
        except (OSError, AttributeError, ImportError):
            return False

        return is_cfrad



"""
Chimera frontend tests

"""

import numpy as np
from numpy.testing import assert_almost_equal, assert_array_equal, assert_equal

from yt.testing import requires_file, requires_module
from yt.utilities.answer_testing.framework import (
    GenericArrayTest,
    data_dir_load,
    requires_ds,
)

Two_D = "F37_80/chimera_00001_grid_1_01.h5"


@requires_module("h5py")
@requires_ds(Two_D)
def test_2D():
    ds = data_dir_load(Two_D)
    _fields = [
        ("chimera", "a_nuc_rep_c"),
        ("chimera", "abar"),
        ("chimera", "ar36"),
        ("chimera", "be_nuc_rep_c"),
        ("chimera", "c12"),
        ("chimera", "ca40"),
        ("chimera", "cr48"),
        ("chimera", "dudt_nu"),
        ("chimera", "dudt_nuc"),
        ("chimera", "e_book"),
        ("chimera", "e_int"),
        ("chimera", "e_rms_1"),
        ("chimera", "e_rms_2"),
        ("chimera", "e_rms_3"),
        ("chimera", "e_rms_4"),
        ("chimera", "entropy"),
        ("chimera", "fe52"),
        ("chimera", "fe56"),
        ("chimera", "grav_x_c"),
        ("chimera", "grav_y_c"),
        ("chimera", "grav_z_c"),
        ("chimera", "he4"),
        ("chimera", "lumin_1"),
        ("chimera", "lumin_2"),
        ("chimera", "lumin_3"),
        ("chimera", "lumin_4"),
        ("chimera", "mg24"),
        ("chimera", "n"),
        ("chimera", "ne20"),
        ("chimera", "ni56"),
        ("chimera", "nse_c"),
        ("chimera", "num_lumin_1"),
        ("chimera", "num_lumin_2"),
        ("chimera", "num_lumin_3"),
        ("chimera", "num_lumin_4"),
        ("chimera", "o16"),
        ("chimera", "p"),
        ("chimera", "press"),
        ("chimera", "rho_c"),
        ("chimera", "s32"),
        ("chimera", "si28"),
        ("chimera", "t_c"),
        ("chimera", "ti44"),
        ("chimera", "u_c"),
        ("chimera", "v_c"),
        ("chimera", "v_csound"),
        ("chimera", "wBVMD"),
        ("chimera", "w_c"),
        ("chimera", "ye_c"),
        ("chimera", "ylep"),
        ("chimera", "z_nuc_rep_c"),
        ("chimera", "zn60"),
    ]
    assert_equal(str(ds), "chimera_00001_grid_1_01.h5")
    assert_equal(str(ds.geometry), "spherical")  # Geometry
    assert_almost_equal(
        ds.domain_right_edge,
        ds.arr([1.0116509e10 + 100, 3.14159265e00, 6.28318531e00], "code_length"),
    )  # domain edge
    assert_array_equal(
        ds.domain_left_edge, ds.arr([0.0, 0.0, 0.0], "code_length")
    )  # domain edge
    assert_array_equal(ds.domain_dimensions, np.array([722, 240, 1]))  # Dimensions
    assert_array_equal(ds.field_list, _fields)

    def field_func(field):
        min = dd[field].min()
        max = dd[field].max()
        avg = np.mean(dd[field])
        size = dd[field].size
        return [min, max, avg, size]

    dd = ds.all_data()
    for field in _fields:
        if field != ("chimera", "shock"):
            yield GenericArrayTest(ds, field_func, args=[field])


Three_D = "C15-3D-3deg/chimera_002715000_grid_1_01.h5"


@requires_module("h5py")
@requires_ds(Three_D)
def test_3D():
    ds = data_dir_load(Three_D)
    _fields = [
        ("chimera", "a_nuc_rep_c"),
        ("chimera", "abar"),
        ("chimera", "ar36"),
        ("chimera", "be_nuc_rep_c"),
        ("chimera", "c12"),
        ("chimera", "ca40"),
        ("chimera", "cr48"),
        ("chimera", "dudt_nu"),
        ("chimera", "dudt_nuc"),
        ("chimera", "e_book"),
        ("chimera", "e_int"),
        ("chimera", "e_rms_1"),
        ("chimera", "e_rms_2"),
        ("chimera", "e_rms_3"),
        ("chimera", "e_rms_4"),
        ("chimera", "entropy"),
        ("chimera", "fe52"),
        ("chimera", "grav_x_c"),
        ("chimera", "grav_y_c"),
        ("chimera", "grav_z_c"),
        ("chimera", "he4"),
        ("chimera", "lumin_1"),
        ("chimera", "lumin_2"),
        ("chimera", "lumin_3"),
        ("chimera", "lumin_4"),
        ("chimera", "mg24"),
        ("chimera", "n"),
        ("chimera", "ne20"),
        ("chimera", "ni56"),
        ("chimera", "nse_c"),
        ("chimera", "num_lumin_1"),
        ("chimera", "num_lumin_2"),
        ("chimera", "num_lumin_3"),
        ("chimera", "num_lumin_4"),
        ("chimera", "o16"),
        ("chimera", "p"),
        ("chimera", "press"),
        ("chimera", "rho_c"),
        ("chimera", "s32"),
        ("chimera", "si28"),
        ("chimera", "t_c"),
        ("chimera", "ti44"),
        ("chimera", "u_c"),
        ("chimera", "v_c"),
        ("chimera", "v_csound"),
        ("chimera", "wBVMD"),
        ("chimera", "w_c"),
        ("chimera", "ye_c"),
        ("chimera", "ylep"),
        ("chimera", "z_nuc_rep_c"),
        ("chimera", "zn60"),
    ]
    assert_equal(str(ds), "chimera_002715000_grid_1_01.h5")
    assert_equal(str(ds.geometry), "spherical")  # Geometry
    assert_almost_equal(
        ds.domain_right_edge,
        ds.arr(
            [1.06500257e09 - 1.03818333, 3.14159265e00, 6.2831853e00], "code_length"
        ),
    )  # Domain edge
    assert_array_equal(ds.domain_left_edge, [0.0, 0.0, 0.0])  # Domain edge
    assert_array_equal(ds.domain_dimensions, [542, 60, 135])  # Dimensions
    assert_array_equal(ds.field_list, _fields)

    def field_func(field):
        min = dd[field].min()
        max = dd[field].max()
        avg = np.mean(dd[field])
        size = dd[field].size
        return [min, max, avg, size]

    dd = ds.all_data()
    for field in _fields:
        if field != ("chimera", "shock"):
            yield GenericArrayTest(ds, field_func, args=[field])


@requires_file(Three_D)
def test_multimesh():  # Tests that the multimesh system for 3D data has been created correctly
    ds = data_dir_load(Three_D)
    assert_equal(len(ds.index.meshes), 45)
    for i in range(44):
        assert_almost_equal(
            ds.index.meshes[i + 1].connectivity_coords
            - ds.index.meshes[i].connectivity_coords,
            np.tile([0.0, 0.0, 0.13962634015954636], (132004, 1)),
        )  # Tests that each mesh is an identically shaped wedge, incrememnted in Phi.
        assert_array_equal(
            ds.index.meshes[i + 1].connectivity_indices,
            ds.index.meshes[i].connectivity_indices,
        )  # Checks Connectivity array is identical for all meshes.



import os
import weakref
from collections import defaultdict
from functools import cached_property
from numbers import Number as numeric_type
from typing import Tuple, Type

import numpy as np

from yt.data_objects.index_subobjects.grid_patch import AMRGridPatch
from yt.data_objects.profiles import (
    Profile1DFromDataset,
    Profile2DFromDataset,
    Profile3DFromDataset,
)
from yt.data_objects.static_output import Dataset, ParticleFile
from yt.data_objects.unions import ParticleUnion
from yt.fields.field_exceptions import NeedsGridType
from yt.fields.field_info_container import FieldInfoContainer
from yt.funcs import is_root, parse_h5_attr
from yt.geometry.api import Geometry
from yt.geometry.geometry_handler import Index
from yt.geometry.grid_geometry_handler import GridIndex
from yt.geometry.particle_geometry_handler import ParticleIndex
from yt.units import dimensions
from yt.units._numpy_wrapper_functions import uconcatenate
from yt.units.unit_registry import UnitRegistry  # type: ignore
from yt.units.yt_array import YTQuantity
from yt.utilities.exceptions import GenerationInProgress, YTFieldTypeNotFound
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.on_demand_imports import _h5py as h5py
from yt.utilities.parallel_tools.parallel_analysis_interface import parallel_root_only
from yt.utilities.tree_container import TreeContainer

from .fields import YTDataContainerFieldInfo, YTGridFieldInfo

_grid_data_containers = ["arbitrary_grid", "covering_grid", "smoothed_covering_grid"]
_set_attrs = {"periodicity": "_periodicity"}


class SavedDataset(Dataset):
    """
    Base dataset class for products of calling save_as_dataset.
    """

    geometry = Geometry.CARTESIAN
    _con_attrs: Tuple[str, ...] = ()

    def _parse_parameter_file(self):
        self.refine_by = 2
        with h5py.File(self.parameter_filename, mode="r") as f:
            for key in f.attrs.keys():
                v = parse_h5_attr(f, key)
                if key == "con_args":
                    try:
                        v = eval(v)
                    except ValueError:
                        # support older ytdata outputs
                        v = v.astype("str")
                    except NameError:
                        # This is the most common error we expect, and it
                        # results from having the eval do a concatenated decoded
                        # set of the values.
                        v = [_.decode("utf8") for _ in v]
                self.parameters[key] = v
            self._with_parameter_file_open(f)

        # if saved, restore unit registry from the json string
        if "unit_registry_json" in self.parameters:
            self.unit_registry = UnitRegistry.from_json(
                self.parameters["unit_registry_json"]
            )
            # reset self.arr and self.quan to use new unit_registry
            self._arr = None
            self._quan = None
            for dim in [
                "length",
                "mass",
                "pressure",
                "temperature",
                "time",
                "velocity",
            ]:
                cu = "code_" + dim
                if cu not in self.unit_registry:
                    self.unit_registry.add(cu, 1.0, getattr(dimensions, dim))
            if "code_magnetic" not in self.unit_registry:
                self.unit_registry.add(
                    "code_magnetic", 0.1**0.5, dimensions.magnetic_field_cgs
                )

        # if saved, set unit system
        if "unit_system_name" in self.parameters:
            unit_system = self.parameters["unit_system_name"]
            del self.parameters["unit_system_name"]
        else:
            unit_system = "cgs"
        # reset unit system since we may have a new unit registry
        self._assign_unit_system(unit_system)

        # assign units to parameters that have associated unit string
        del_pars = []
        for par in self.parameters:
            ustr = f"{par}_units"
            if ustr in self.parameters:
                if isinstance(self.parameters[par], np.ndarray):
                    to_u = self.arr
                else:
                    to_u = self.quan
                self.parameters[par] = to_u(self.parameters[par], self.parameters[ustr])
                del_pars.append(ustr)
        for par in del_pars:
            del self.parameters[par]

        for attr in self._con_attrs:
            sattr = _set_attrs.get(attr, attr)
            if sattr == "geometry":
                if "geometry" in self.parameters:
                    self.geometry = Geometry(self.parameters["geometry"])
                continue
            try:
                setattr(self, sattr, self.parameters.get(attr))
            except TypeError:
                # some Dataset attributes are properties with setters
                # which may not accept None as an input
                pass

    def _with_parameter_file_open(self, f):
        # This allows subclasses to access the parameter file
        # while it's still open to get additional information.
        pass

    def set_units(self):
        if "unit_registry_json" in self.parameters:
            self._set_code_unit_attributes()
            del self.parameters["unit_registry_json"]
        else:
            super().set_units()

    def _set_code_unit_attributes(self):
        attrs = (
            "length_unit",
            "mass_unit",
            "time_unit",
            "velocity_unit",
            "magnetic_unit",
        )
        cgs_units = ("cm", "g", "s", "cm/s", "gauss")
        base_units = np.ones(len(attrs))
        for unit, attr, cgs_unit in zip(base_units, attrs, cgs_units):
            if attr in self.parameters and isinstance(
                self.parameters[attr], YTQuantity
            ):
                uq = self.parameters[attr]
            elif attr in self.parameters and f"{attr}_units" in self.parameters:
                uq = self.quan(self.parameters[attr], self.parameters[f"{attr}_units"])
                del self.parameters[attr]
                del self.parameters[f"{attr}_units"]
            elif isinstance(unit, str):
                uq = self.quan(1.0, unit)
            elif isinstance(unit, numeric_type):
                uq = self.quan(unit, cgs_unit)
            elif isinstance(unit, YTQuantity):
                uq = unit
            elif isinstance(unit, tuple):
                uq = self.quan(unit[0], unit[1])
            else:
                raise RuntimeError(f"{attr} ({unit}) is invalid.")
            setattr(self, attr, uq)


class YTDataset(SavedDataset):
    """Base dataset class for all ytdata datasets."""

    _con_attrs = (
        "cosmological_simulation",
        "current_time",
        "current_redshift",
        "hubble_constant",
        "omega_matter",
        "omega_lambda",
        "dimensionality",
        "domain_dimensions",
        "geometry",
        "periodicity",
        "domain_left_edge",
        "domain_right_edge",
        "container_type",
        "data_type",
    )

    def _with_parameter_file_open(self, f):
        self.num_particles = {
            group: parse_h5_attr(f[group], "num_elements")
            for group in f
            if group != self.default_fluid_type
        }

    def create_field_info(self):
        self.field_dependencies = {}
        self.derived_field_list = []
        self.filtered_particle_types = []
        self.field_info = self._field_info_class(self, self.field_list)
        self.coordinates.setup_fields(self.field_info)
        self.field_info.setup_fluid_fields()
        for ptype in self.particle_types:
            self.field_info.setup_particle_fields(ptype)

        self._setup_gas_alias()
        self.field_info.setup_fluid_index_fields()

        if "all" not in self.particle_types:
            mylog.debug("Creating Particle Union 'all'")
            pu = ParticleUnion("all", list(self.particle_types_raw))
            self.add_particle_union(pu)
        self.field_info.setup_extra_union_fields()
        mylog.debug("Loading field plugins.")
        self.field_info.load_all_plugins()
        deps, unloaded = self.field_info.check_derived_fields()
        self.field_dependencies.update(deps)

    def _setup_gas_alias(self):
        pass

    def _setup_override_fields(self):
        pass


class YTDataHDF5File(ParticleFile):
    def __init__(self, ds, io, filename, file_id, range):
        with h5py.File(filename, mode="r") as f:
            self.header = {field: parse_h5_attr(f, field) for field in f.attrs.keys()}

        super().__init__(ds, io, filename, file_id, range)


class YTDataContainerDataset(YTDataset):
    """Dataset for saved geometric data containers."""

    _index_class = ParticleIndex
    _file_class = YTDataHDF5File
    _field_info_class: Type[FieldInfoContainer] = YTDataContainerFieldInfo
    _suffix = ".h5"
    fluid_types = ("grid", "gas", "deposit", "index")

    def __init__(
        self,
        filename,
        dataset_type="ytdatacontainer_hdf5",
        index_order=None,
        index_filename=None,
        units_override=None,
        unit_system="cgs",
    ):
        self.index_order = index_order
        self.index_filename = index_filename
        super().__init__(
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
        )

    def _parse_parameter_file(self):
        super()._parse_parameter_file()
        self.particle_types_raw = tuple(self.num_particles.keys())
        self.particle_types = self.particle_types_raw
        self.filename_template = self.parameter_filename
        self.file_count = 1
        self.domain_dimensions = np.ones(3, "int32")

    def _setup_gas_alias(self):
        "Alias the grid type to gas by making a particle union."

        if "grid" in self.particle_types and "gas" not in self.particle_types:
            pu = ParticleUnion("gas", ["grid"])
            self.add_particle_union(pu)
        # We have to alias this because particle unions only
        # cover the field_list.
        self.field_info.alias(("gas", "cell_volume"), ("grid", "cell_volume"))

    @cached_property
    def data(self):
        """
        Return a data container configured like the original used to
        create this dataset.
        """

        # Some data containers can't be reconstructed in the same way
        # since this is now particle-like data.
        data_type = self.parameters.get("data_type")
        container_type = self.parameters.get("container_type")
        ex_container_type = ["cutting", "quad_proj", "ray", "slice", "cut_region"]
        if data_type == "yt_light_ray" or container_type in ex_container_type:
            mylog.info("Returning an all_data data container.")
            return self.all_data()

        my_obj = getattr(self, self.parameters["container_type"])
        my_args = [self.parameters[con_arg] for con_arg in self.parameters["con_args"]]
        return my_obj(*my_args)

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        if not filename.endswith(".h5"):
            return False
        with h5py.File(filename, mode="r") as f:
            data_type = parse_h5_attr(f, "data_type")
            cont_type = parse_h5_attr(f, "container_type")
            if data_type is None:
                return False
            if (
                data_type == "yt_data_container"
                and cont_type not in _grid_data_containers
            ):
                return True
        return False


class YTDataLightRayDataset(YTDataContainerDataset):
    """Dataset for saved LightRay objects."""

    def _parse_parameter_file(self):
        super()._parse_parameter_file()
        self._restore_light_ray_solution()

    def _restore_light_ray_solution(self):
        """
        Restore all information associate with the light ray solution
        to its original form.
        """
        key = "light_ray_solution"
        self.light_ray_solution = []
        lrs_fields = [
            par for par in self.parameters if key in par and not par.endswith("_units")
        ]
        if len(lrs_fields) == 0:
            return
        self.light_ray_solution = [{} for val in self.parameters[lrs_fields[0]]]
        for sp3 in ["unique_identifier", "filename"]:
            ksp3 = f"{key}_{sp3}"
            if ksp3 not in lrs_fields:
                continue
            self.parameters[ksp3] = self.parameters[ksp3].astype(str)
        for field in lrs_fields:
            field_name = field[len(key) + 1 :]
            for i in range(self.parameters[field].shape[0]):
                self.light_ray_solution[i][field_name] = self.parameters[field][i]

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        if not filename.endswith(".h5"):
            return False
        with h5py.File(filename, mode="r") as f:
            data_type = parse_h5_attr(f, "data_type")
            if data_type in ["yt_light_ray"]:
                return True
        return False


class YTSpatialPlotDataset(YTDataContainerDataset):
    """Dataset for saved slices and projections."""

    _field_info_class = YTGridFieldInfo

    def __init__(self, *args, **kwargs):
        super().__init__(*args, dataset_type="ytspatialplot_hdf5", **kwargs)

    def _parse_parameter_file(self):
        super()._parse_parameter_file()
        if self.parameters["container_type"] == "proj":
            if (
                isinstance(self.parameters["weight_field"], str)
                and self.parameters["weight_field"] == "None"
            ):
                self.parameters["weight_field"] = None
            elif isinstance(self.parameters["weight_field"], np.ndarray):
                self.parameters["weight_field"] = tuple(self.parameters["weight_field"])

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        if not filename.endswith(".h5"):
            return False
        with h5py.File(filename, mode="r") as f:
            data_type = parse_h5_attr(f, "data_type")
            cont_type = parse_h5_attr(f, "container_type")
            if data_type == "yt_data_container" and cont_type in [
                "cutting",
                "proj",
                "slice",
                "quad_proj",
            ]:
                return True
        return False


class YTGrid(AMRGridPatch):
    _id_offset = 0

    def __init__(self, gid, index, filename=None):
        AMRGridPatch.__init__(self, gid, filename=filename, index=index)
        self._children_ids = []
        self._parent_id = -1
        self.Level = 0
        self.LeftEdge = self.index.ds.domain_left_edge
        self.RightEdge = self.index.ds.domain_right_edge

    def __getitem__(self, key):
        tr = super(AMRGridPatch, self).__getitem__(key)
        try:
            fields = self._determine_fields(key)
        except YTFieldTypeNotFound:
            return tr
        finfo = self.ds._get_field_info(fields[0])
        if not finfo.sampling_type == "particle":
            return tr.reshape(self.ActiveDimensions[: self.ds.dimensionality])
        return tr

    @property
    def Parent(self):
        return None

    @property
    def Children(self):
        return []


class YTDataHierarchy(GridIndex):
    def __init__(self, ds, dataset_type=None):
        self.dataset_type = dataset_type
        self.float_type = "float64"
        self.dataset = weakref.proxy(ds)
        self.directory = os.getcwd()
        super().__init__(ds, dataset_type)

    def _count_grids(self):
        self.num_grids = 1

    def _parse_index(self):
        self.grid_dimensions[:] = self.ds.domain_dimensions
        self.grid_left_edge[:] = self.ds.domain_left_edge
        self.grid_right_edge[:] = self.ds.domain_right_edge
        self.grid_levels[:] = np.zeros(self.num_grids)
        self.grid_procs = np.zeros(self.num_grids)
        self.grid_particle_count[:] = sum(self.ds.num_particles.values())
        self.grids = []
        for gid in range(self.num_grids):
            self.grids.append(self.grid(gid, self))
            self.grids[gid].Level = self.grid_levels[gid, 0]
        self.max_level = self.grid_levels.max()
        temp_grids = np.empty(self.num_grids, dtype="object")
        for i, grid in enumerate(self.grids):
            grid.filename = self.ds.parameter_filename
            grid._prepare_grid()
            grid.proc_num = self.grid_procs[i]
            temp_grids[i] = grid
        self.grids = temp_grids

    def _detect_output_fields(self):
        self.field_list = []
        self.ds.field_units = self.ds.field_units or {}
        with h5py.File(self.ds.parameter_filename, mode="r") as f:
            for group in f:
                for field in f[group]:
                    field_name = (str(group), str(field))
                    self.field_list.append(field_name)
                    self.ds.field_units[field_name] = parse_h5_attr(
                        f[group][field], "units"
                    )


class YTGridHierarchy(YTDataHierarchy):
    grid = YTGrid

    def _populate_grid_objects(self):
        for g in self.grids:
            g._setup_dx()
        self.max_level = self.grid_levels.max()


class YTGridDataset(YTDataset):
    """Dataset for saved covering grids, arbitrary grids, and FRBs."""

    _index_class: Type[Index] = YTGridHierarchy
    _field_info_class = YTGridFieldInfo
    _dataset_type = "ytgridhdf5"
    geometry = Geometry.CARTESIAN
    default_fluid_type = "grid"
    fluid_types: Tuple[str, ...] = ("grid", "gas", "deposit", "index")

    def __init__(self, filename, unit_system="cgs"):
        super().__init__(filename, self._dataset_type, unit_system=unit_system)
        self.data = self.index.grids[0]

    def _parse_parameter_file(self):
        super()._parse_parameter_file()
        self.num_particles.pop(self.default_fluid_type, None)
        self.particle_types_raw = tuple(self.num_particles.keys())
        self.particle_types = self.particle_types_raw

        # correct domain dimensions for the covering grid dimension
        self.base_domain_left_edge = self.domain_left_edge
        self.base_domain_right_edge = self.domain_right_edge
        self.base_domain_dimensions = self.domain_dimensions

        if self.container_type in _grid_data_containers:
            self.domain_left_edge = self.parameters["left_edge"]

            if "level" in self.parameters["con_args"]:
                dx = (self.base_domain_right_edge - self.base_domain_left_edge) / (
                    self.domain_dimensions * self.refine_by ** self.parameters["level"]
                )
                self.domain_right_edge = (
                    self.domain_left_edge + self.parameters["ActiveDimensions"] * dx
                )
                self.domain_dimensions = (
                    (self.domain_right_edge - self.domain_left_edge) / dx
                ).astype(int)
            else:
                self.domain_right_edge = self.parameters["right_edge"]
                self.domain_dimensions = self.parameters["ActiveDimensions"]
                dx = (
                    self.domain_right_edge - self.domain_left_edge
                ) / self.domain_dimensions

            periodicity = (
                np.abs(self.domain_left_edge - self.base_domain_left_edge) < 0.5 * dx
            )
            periodicity &= (
                np.abs(self.domain_right_edge - self.base_domain_right_edge) < 0.5 * dx
            )
            self._periodicity = periodicity

        elif self.data_type == "yt_frb":
            dle = self.domain_left_edge
            self.domain_left_edge = uconcatenate(
                [self.parameters["left_edge"].to(dle.units), [0] * dle.uq]
            )
            dre = self.domain_right_edge
            self.domain_right_edge = uconcatenate(
                [self.parameters["right_edge"].to(dre.units), [1] * dre.uq]
            )
            self.domain_dimensions = np.concatenate(
                [self.parameters["ActiveDimensions"], [1]]
            )

    def _setup_gas_alias(self):
        "Alias the grid type to gas with a field alias."

        for ftype, field in self.field_list:
            if ftype == "grid":
                self.field_info.alias(("gas", field), ("grid", field))

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        if not filename.endswith(".h5"):
            return False
        with h5py.File(filename, mode="r") as f:
            data_type = parse_h5_attr(f, "data_type")
            cont_type = parse_h5_attr(f, "container_type")
            if data_type == "yt_frb":
                return True
            if data_type == "yt_data_container" and cont_type in _grid_data_containers:
                return True
        return False


class YTNonspatialGrid(AMRGridPatch):
    _id_offset = 0

    def __init__(self, gid, index, filename=None):
        super().__init__(gid, filename=filename, index=index)
        self._children_ids = []
        self._parent_id = -1
        self.Level = 0
        self.LeftEdge = self.index.ds.domain_left_edge
        self.RightEdge = self.index.ds.domain_right_edge

    def __getitem__(self, key):
        tr = super(AMRGridPatch, self).__getitem__(key)
        try:
            fields = self._determine_fields(key)
        except YTFieldTypeNotFound:
            return tr
        self.ds._get_field_info(fields[0])
        return tr

    def get_data(self, fields=None):
        if fields is None:
            return
        nfields = []
        apply_fields = defaultdict(list)
        for field in self._determine_fields(fields):
            if field[0] in self.ds.filtered_particle_types:
                f = self.ds.known_filters[field[0]]
                apply_fields[field[0]].append((f.filtered_type, field[1]))
            else:
                nfields.append(field)
        for filter_type in apply_fields:
            f = self.ds.known_filters[filter_type]
            with f.apply(self):
                self.get_data(apply_fields[filter_type])
        fields = nfields
        if len(fields) == 0:
            return
        # Now we collect all our fields
        # Here is where we need to perform a validation step, so that if we
        # have a field requested that we actually *can't* yet get, we put it
        # off until the end.  This prevents double-reading fields that will
        # need to be used in spatial fields later on.
        fields_to_get = []
        # This will be pre-populated with spatial fields
        fields_to_generate = []
        for field in self._determine_fields(fields):
            if field in self.field_data:
                continue
            finfo = self.ds._get_field_info(field)
            try:
                finfo.check_available(self)
            except NeedsGridType:
                fields_to_generate.append(field)
                continue
            fields_to_get.append(field)
        if len(fields_to_get) == 0 and len(fields_to_generate) == 0:
            return
        elif self._locked:
            raise GenerationInProgress(fields)
        # Track which ones we want in the end
        ofields = set(list(self.field_data.keys()) + fields_to_get + fields_to_generate)
        # At this point, we want to figure out *all* our dependencies.
        fields_to_get = self._identify_dependencies(fields_to_get, self._spatial)
        # We now split up into readers for the types of fields
        fluids, particles = [], []
        finfos = {}
        for field_key in fields_to_get:
            finfo = self.ds._get_field_info(field_key)
            finfos[field_key] = finfo
            if finfo.sampling_type == "particle":
                particles.append(field_key)
            elif field_key not in fluids:
                fluids.append(field_key)

        # The _read method will figure out which fields it needs to get from
        # disk, and return a dict of those fields along with the fields that
        # need to be generated.
        read_fluids, gen_fluids = self.index._read_fluid_fields(
            fluids, self, self._current_chunk
        )
        for f, v in read_fluids.items():
            convert = True
            if v.dtype != np.float64:
                if finfos[f].units == "":
                    self.field_data[f] = v
                    convert = False
                else:
                    v = v.astype(np.float64)
            if convert:
                self.field_data[f] = self.ds.arr(v, units=finfos[f].units)
                self.field_data[f].convert_to_units(finfos[f].output_units)

        read_particles, gen_particles = self.index._read_fluid_fields(
            particles, self, self._current_chunk
        )
        for f, v in read_particles.items():
            convert = True
            if v.dtype != np.float64:
                if finfos[f].units == "":
                    self.field_data[f] = v
                    convert = False
                else:
                    v = v.astype(np.float64)
            if convert:
                self.field_data[f] = self.ds.arr(v, units=finfos[f].units)
                self.field_data[f].convert_to_units(finfos[f].output_units)

        fields_to_generate += gen_fluids + gen_particles
        self._generate_fields(fields_to_generate)
        for field in list(self.field_data.keys()):
            if field not in ofields:
                self.field_data.pop(field)

    @property
    def Parent(self):
        return None

    @property
    def Children(self):
        return []


class YTNonspatialHierarchy(YTDataHierarchy):
    grid = YTNonspatialGrid

    def _populate_grid_objects(self):
        for g in self.grids:
            g._setup_dx()
            # this is non-spatial, so remove the code_length units
            g.dds = self.ds.arr(g.dds.d, "")
            g.ActiveDimensions = self.ds.domain_dimensions
        self.max_level = self.grid_levels.max()

    def _read_fluid_fields(self, fields, dobj, chunk=None):
        if len(fields) == 0:
            return {}, []
        fields_to_read, fields_to_generate = self._split_fields(fields)
        if len(fields_to_read) == 0:
            return {}, fields_to_generate
        selector = dobj.selector
        fields_to_return = self.io._read_fluid_selection(dobj, selector, fields_to_read)
        return fields_to_return, fields_to_generate


class YTNonspatialDataset(YTGridDataset):
    """Dataset for general array data."""

    _index_class = YTNonspatialHierarchy
    _field_info_class = YTGridFieldInfo
    _dataset_type = "ytnonspatialhdf5"
    geometry = Geometry.CARTESIAN
    default_fluid_type = "data"
    fluid_types: Tuple[str, ...] = ("data", "gas")

    def _parse_parameter_file(self):
        super(YTGridDataset, self)._parse_parameter_file()
        self.num_particles.pop(self.default_fluid_type, None)
        self.particle_types_raw = tuple(self.num_particles.keys())
        self.particle_types = self.particle_types_raw

    def _set_derived_attrs(self):
        # set some defaults just to make things go
        default_attrs = {
            "dimensionality": 3,
            "domain_dimensions": np.ones(3, dtype="int64"),
            "domain_left_edge": np.zeros(3),
            "domain_right_edge": np.ones(3),
            "_periodicity": np.ones(3, dtype="bool"),
        }
        for att, val in default_attrs.items():
            if getattr(self, att, None) is None:
                setattr(self, att, val)

    def _setup_classes(self):
        # We don't allow geometric selection for non-spatial datasets
        self.objects = []

    @parallel_root_only
    def print_key_parameters(self):
        for a in [
            "current_time",
            "domain_dimensions",
            "domain_left_edge",
            "domain_right_edge",
            "cosmological_simulation",
        ]:
            v = getattr(self, a)
            if v is not None:
                mylog.info("Parameters: %-25s = %s", a, v)
        if hasattr(self, "cosmological_simulation") and self.cosmological_simulation:
            for a in [
                "current_redshift",
                "omega_lambda",
                "omega_matter",
                "hubble_constant",
            ]:
                v = getattr(self, a)
                if v is not None:
                    mylog.info("Parameters: %-25s = %s", a, v)
        mylog.warning("Geometric data selection not available for this dataset type.")

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        if not filename.endswith(".h5"):
            return False
        with h5py.File(filename, mode="r") as f:
            data_type = parse_h5_attr(f, "data_type")
            if data_type == "yt_array_data":
                return True
        return False


class YTProfileDataset(YTNonspatialDataset):
    """Dataset for saved profile objects."""

    fluid_types = ("data", "gas", "standard_deviation")

    def __init__(self, filename, unit_system="cgs"):
        super().__init__(filename, unit_system=unit_system)

    _profile = None

    @property
    def profile(self):
        if self._profile is not None:
            return self._profile
        if self.dimensionality == 1:
            self._profile = Profile1DFromDataset(self)
        elif self.dimensionality == 2:
            self._profile = Profile2DFromDataset(self)
        elif self.dimensionality == 3:
            self._profile = Profile3DFromDataset(self)
        else:
            self._profile = None
        return self._profile

    def _parse_parameter_file(self):
        super(YTGridDataset, self)._parse_parameter_file()

        if (
            isinstance(self.parameters["weight_field"], str)
            and self.parameters["weight_field"] == "None"
        ):
            self.parameters["weight_field"] = None
        elif isinstance(self.parameters["weight_field"], np.ndarray):
            self.parameters["weight_field"] = tuple(
                self.parameters["weight_field"].astype(str)
            )

        for a in ["profile_dimensions"] + [
            f"{ax}_{attr}" for ax in "xyz"[: self.dimensionality] for attr in ["log"]
        ]:
            setattr(self, a, self.parameters[a])

        self.base_domain_left_edge = self.domain_left_edge
        self.base_domain_right_edge = self.domain_right_edge
        self.base_domain_dimensions = self.domain_dimensions

        domain_dimensions = np.ones(3, dtype="int64")
        domain_dimensions[: self.dimensionality] = self.profile_dimensions
        self.domain_dimensions = domain_dimensions
        domain_left_edge = np.zeros(3)
        domain_right_edge = np.ones(3)
        for i, ax in enumerate("xyz"[: self.dimensionality]):
            range_name = f"{ax}_range"
            my_range = self.parameters[range_name]
            if getattr(self, f"{ax}_log", False):
                my_range = np.log10(my_range)
            domain_left_edge[i] = my_range[0]
            domain_right_edge[i] = my_range[1]
            setattr(self, range_name, self.parameters[range_name])

            bin_field = f"{ax}_field"
            if (
                isinstance(self.parameters[bin_field], str)
                and self.parameters[bin_field] == "None"
            ):
                self.parameters[bin_field] = None
            elif isinstance(self.parameters[bin_field], np.ndarray):
                self.parameters[bin_field] = (
                    "data",
                    self.parameters[bin_field].astype(str)[1],
                )
            setattr(self, bin_field, self.parameters[bin_field])
        self.domain_left_edge = domain_left_edge
        self.domain_right_edge = domain_right_edge

    def _setup_gas_alias(self):
        "Alias the grid type to gas with a field alias."
        for ftype, field in self.field_list:
            if ftype == "data":
                self.field_info.alias(("gas", field), (ftype, field))

    def create_field_info(self):
        super().create_field_info()
        if self.parameters["weight_field"] is not None:
            self.field_info.alias(
                self.parameters["weight_field"], (self.default_fluid_type, "weight")
            )

    def _set_derived_attrs(self):
        self.domain_center = 0.5 * (self.domain_right_edge + self.domain_left_edge)
        self.domain_width = self.domain_right_edge - self.domain_left_edge

    def print_key_parameters(self):
        if is_root():
            mylog.info("YTProfileDataset")
            for a in ["dimensionality", "profile_dimensions"] + [
                f"{ax}_{attr}"
                for ax in "xyz"[: self.dimensionality]
                for attr in ["field", "range", "log"]
            ]:
                v = getattr(self, a)
                mylog.info("Parameters: %-25s = %s", a, v)
        super().print_key_parameters()

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        if not filename.endswith(".h5"):
            return False
        with h5py.File(filename, mode="r") as f:
            data_type = parse_h5_attr(f, "data_type")
            if data_type == "yt_profile":
                return True
        return False


class YTClumpContainer(TreeContainer):
    def __init__(
        self, clump_id, global_id, parent_id, contour_key, contour_id, ds=None
    ):
        self.clump_id = clump_id
        self.global_id = global_id
        self.parent_id = parent_id
        self.contour_key = contour_key
        self.contour_id = contour_id
        self.parent = None
        self.ds = ds
        TreeContainer.__init__(self)

    def add_child(self, child):
        if self.children is None:
            self.children = []
        self.children.append(child)
        child.parent = self

    def __repr__(self):
        return "Clump[%d]" % self.clump_id

    def __getitem__(self, field):
        g = self.ds.data
        f = g._determine_fields(field)[0]
        if f[0] == "clump":
            return g[f][self.global_id]
        if self.contour_id == -1:
            return g[f]
        cfield = (f[0], f"contours_{self.contour_key.decode('utf-8')}")
        if f[0] == "grid":
            return g[f][g[cfield] == self.contour_id]
        return self.parent[f][g[cfield] == self.contour_id]


class YTClumpTreeDataset(YTNonspatialDataset):
    """Dataset for saved clump-finder data."""

    def __init__(self, filename, unit_system="cgs"):
        super().__init__(filename, unit_system=unit_system)
        self._load_tree()

    def _load_tree(self):
        my_tree = {}
        for i, clump_id in enumerate(self.data[("clump", "clump_id")]):
            my_tree[clump_id] = YTClumpContainer(
                clump_id,
                i,
                self.data["clump", "parent_id"][i],
                self.data["clump", "contour_key"][i],
                self.data["clump", "contour_id"][i],
                self,
            )
        for clump in my_tree.values():
            if clump.parent_id == -1:
                self.tree = clump
            else:
                parent = my_tree[clump.parent_id]
                parent.add_child(clump)

    @cached_property
    def leaves(self):
        return [clump for clump in self.tree if clump.children is None]

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        if not filename.endswith(".h5"):
            return False
        with h5py.File(filename, mode="r") as f:
            data_type = parse_h5_attr(f, "data_type")
            if data_type is None:
                return False
            if data_type == "yt_clump_tree":
                return True
        return False



# Some tests for the geographic coordinates handler

import numpy as np
from numpy.testing import assert_equal

from yt.testing import assert_rel_equal, fake_amr_ds

# Our canonical tests are that we can access all of our fields and we can
# compute our volume correctly.


def test_geographic_coordinates():
    # We're going to load up a simple AMR grid and check its volume
    # calculations and path length calculations.

    # Note that we are setting it up to have an altitude of 1000 maximum, which
    # means our volume will be that of a shell 1000 wide, starting at r of
    # whatever our surface_height is set to.
    ds = fake_amr_ds(geometry="geographic")
    ds.surface_height = ds.quan(5000.0, "code_length")
    axes = ["latitude", "longitude", "altitude"]
    for i, axis in enumerate(axes):
        dd = ds.all_data()
        fi = ("index", axis)
        fd = ("index", f"d{axis}")
        ma = np.argmax(dd[fi])
        assert_equal(dd[fi][ma] + dd[fd][ma] / 2.0, ds.domain_right_edge[i].d)
        mi = np.argmin(dd[fi])
        assert_equal(dd[fi][mi] - dd[fd][mi] / 2.0, ds.domain_left_edge[i].d)
        assert_equal(dd[fd].max(), (ds.domain_width / ds.domain_dimensions)[i].d)
    inner_r = ds.surface_height
    outer_r = ds.surface_height + ds.domain_width[2]
    assert_equal(dd["index", "dtheta"], dd["index", "dlatitude"] * np.pi / 180.0)
    assert_equal(dd["index", "dphi"], dd["index", "dlongitude"] * np.pi / 180.0)
    # Note our terrible agreement here.
    assert_rel_equal(
        dd[("index", "cell_volume")].sum(dtype="float64"),
        (4.0 / 3.0) * np.pi * (outer_r**3 - inner_r**3),
        10,
    )
    assert_equal(dd["index", "path_element_altitude"], dd["index", "daltitude"])
    assert_equal(dd["index", "path_element_altitude"], dd["index", "dr"])
    # Note that latitude corresponds to theta, longitude to phi
    assert_equal(
        dd["index", "path_element_latitude"],
        dd["index", "r"] * dd["index", "dlatitude"] * np.pi / 180.0,
    )
    assert_equal(
        dd["index", "path_element_longitude"],
        (
            dd["index", "r"]
            * dd["index", "dlongitude"]
            * np.pi
            / 180.0
            * np.sin((dd["index", "latitude"] + 90.0) * np.pi / 180.0)
        ),
    )
    # We also want to check that our radius is correct
    assert_equal(dd["index", "r"], dd["index", "altitude"] + ds.surface_height)


def test_internal_geographic_coordinates():
    # We're going to load up a simple AMR grid and check its volume
    # calculations and path length calculations.

    # Note that we are setting it up to have depth of 1000 maximum, which
    # means our volume will be that of a shell 1000 wide, starting at r of
    # outer_radius - 1000.
    ds = fake_amr_ds(geometry="internal_geographic")
    ds.outer_radius = ds.quan(5000.0, "code_length")
    axes = ["latitude", "longitude", "depth"]
    for i, axis in enumerate(axes):
        dd = ds.all_data()
        fi = ("index", axis)
        fd = ("index", f"d{axis}")
        ma = np.argmax(dd[fi])
        assert_equal(dd[fi][ma] + dd[fd][ma] / 2.0, ds.domain_right_edge[i].d)
        mi = np.argmin(dd[fi])
        assert_equal(dd[fi][mi] - dd[fd][mi] / 2.0, ds.domain_left_edge[i].d)
        assert_equal(dd[fd].max(), (ds.domain_width / ds.domain_dimensions)[i].d)
    inner_r = ds.outer_radius - ds.domain_right_edge[2]
    outer_r = ds.outer_radius
    assert_equal(dd["index", "dtheta"], dd["index", "dlatitude"] * np.pi / 180.0)
    assert_equal(dd["index", "dphi"], dd["index", "dlongitude"] * np.pi / 180.0)
    assert_rel_equal(
        dd[("index", "cell_volume")].sum(dtype="float64"),
        (4.0 / 3.0) * np.pi * (outer_r**3 - inner_r**3),
        10,
    )
    assert_equal(dd["index", "path_element_depth"], dd["index", "ddepth"])
    assert_equal(dd["index", "path_element_depth"], dd["index", "dr"])
    # Note that latitude corresponds to theta, longitude to phi
    assert_equal(
        dd["index", "path_element_latitude"],
        dd["index", "r"] * dd["index", "dlatitude"] * np.pi / 180.0,
    )
    assert_equal(
        dd["index", "path_element_longitude"],
        (
            dd["index", "r"]
            * dd["index", "dlongitude"]
            * np.pi
            / 180.0
            * np.sin((dd["index", "latitude"] + 90.0) * np.pi / 180.0)
        ),
    )
    # We also want to check that our radius is correct
    assert_equal(dd["index", "r"], -1.0 * dd["index", "depth"] + ds.outer_radius)



import glob
import os

import numpy as np
from numpy.testing import assert_raises

from yt.config import ytcfg
from yt.data_objects.particle_filters import particle_filter
from yt.data_objects.time_series import DatasetSeries
from yt.testing import fake_particle_ds
from yt.utilities.answer_testing.framework import GenericArrayTest, requires_ds
from yt.utilities.exceptions import YTIllDefinedParticleData


def setup():
    ytcfg["yt", "internals", "within_testing"] = True


data_path = ytcfg.get("yt", "test_data_dir")

pfields = [
    ("all", "particle_position_x"),
    ("all", "particle_position_y"),
    ("all", "particle_position_z"),
]
vfields = [
    ("all", "particle_velocity_x"),
    ("all", "particle_velocity_y"),
    ("all", "particle_velocity_z"),
]


@requires_ds("Orbit/orbit_hdf5_chk_0000")
def test_orbit_traj():
    fields = ["particle_velocity_x", "particle_velocity_y", "particle_velocity_z"]
    my_fns = glob.glob(os.path.join(data_path, "Orbit/orbit_hdf5_chk_00[0-9][0-9]"))
    my_fns.sort()
    ts = DatasetSeries(my_fns)
    ds = ts[0]
    traj = ts.particle_trajectories([1, 2], fields=fields, suppress_logging=True)
    for field in pfields + vfields:

        def field_func(name):
            return traj[field]  # noqa: B023

        yield GenericArrayTest(ds, field_func, args=[field])


@requires_ds("enzo_tiny_cosmology/DD0000/DD0000")
def test_etc_traj():
    fields = [
        ("all", "particle_velocity_x"),
        ("all", "particle_velocity_y"),
        ("all", "particle_velocity_z"),
    ]
    my_fns = glob.glob(
        os.path.join(data_path, "enzo_tiny_cosmology/DD000[0-9]/*.hierarchy")
    )
    my_fns.sort()
    ts = DatasetSeries(my_fns)
    ds = ts[0]
    sp = ds.sphere("max", (0.5, "Mpc"))
    indices = sp["particle_index"][sp["particle_type"] == 1][:5]
    traj = ts.particle_trajectories(indices, fields=fields, suppress_logging=True)
    traj.add_fields([("gas", "density")])
    for field in pfields + vfields + [("gas", "density")]:

        def field_func(name):
            return traj[field]  # noqa: B023

        yield GenericArrayTest(ds, field_func, args=[field])


def test_uniqueness():
    n_particles = 2
    n_steps = 2
    ids = np.arange(n_particles, dtype="int64") % (n_particles // 2)
    data = {"particle_index": ids}
    fields = [
        "particle_position_x",
        "particle_position_y",
        "particle_position_z",
        "particle_index",
    ]
    negative = [False, False, False, False]
    units = ["cm", "cm", "cm", "1"]

    ts = DatasetSeries(
        [
            fake_particle_ds(
                fields=fields,
                negative=negative,
                units=units,
                npart=n_particles,
                data=data,
            )
            for i in range(n_steps)
        ]
    )

    assert_raises(YTIllDefinedParticleData, ts.particle_trajectories, [0])


def test_ptype():
    n_particles = 100
    fields = [
        "particle_position_x",
        "particle_position_y",
        "particle_position_z",
        "particle_index",
        "particle_dummy",
    ]
    negative = [False, False, False, False, False]
    units = ["cm", "cm", "cm", "1", "1"]

    # Setup filters on the 'particle_dummy' field, keeping only the first 50
    @particle_filter(name="dummy", requires=["particle_dummy"])
    def dummy(pfilter, data):
        return data[(pfilter.filtered_type, "particle_dummy")] <= n_particles // 2

    # Setup fake particle datasets with repeated ids. This should work because
    # the ids are unique among `dummy_particles` so let's test this
    data = {
        "particle_index": np.arange(n_particles) % (n_particles // 2),
        "particle_dummy": np.arange(n_particles),
    }
    all_ds = [
        fake_particle_ds(
            fields=fields, negative=negative, units=units, npart=n_particles, data=data
        )
    ]
    for ds in all_ds:
        ds.add_particle_filter("dummy")
    ts = DatasetSeries(all_ds)

    # Select all dummy particles
    print(ts[0].derived_field_list)
    ids = ts[0].all_data()["dummy", "particle_index"]

    # Build trajectories
    ts.particle_trajectories(ids, ptype="dummy")



"""
Geometry selection routine imports.




"""


cimport numpy as np

from yt.geometry.grid_visitors cimport (
    GridTreeNode,
    GridVisitorData,
    check_child_masked,
    grid_visitor_function,
)
from yt.utilities.lib.fp_utils cimport _ensure_code
from yt.utilities.lib.geometry_utils cimport decode_morton_64bit

from .oct_container cimport OctreeContainer
from .oct_visitors cimport Oct, OctVisitor


cdef class SelectorObject:
    cdef public np.int32_t min_level
    cdef public np.int32_t max_level
    cdef public int overlap_cells
    cdef public np.float64_t domain_width[3]
    cdef public np.float64_t domain_center[3]
    cdef public bint periodicity[3]
    cdef bint _hash_initialized
    cdef np.int64_t _hash

    cdef void recursively_visit_octs(self, Oct *root,
                        np.float64_t pos[3], np.float64_t dds[3],
                        int level,
                        OctVisitor visitor,
                        int visit_covered = ?)
    cdef void visit_oct_cells(self, Oct *root, Oct *ch,
                              np.float64_t spos[3], np.float64_t sdds[3],
                              OctVisitor visitor, int i, int j, int k)
    cdef int select_grid(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3],
                               np.int32_t level, Oct *o = ?) noexcept nogil
    cdef int select_grid_edge(self, np.float64_t left_edge[3],
                                    np.float64_t right_edge[3],
                                    np.int32_t level, Oct *o = ?) noexcept nogil
    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil

    cdef int select_point(self, np.float64_t pos[3]) noexcept nogil
    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil
    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil
    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil
    cdef int fill_mask_selector_regular_grid(self, np.float64_t left_edge[3],
                                             np.float64_t right_edge[3],
                                             np.float64_t dds[3], int dim[3],
                                             np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask,
                                             np.ndarray[np.uint8_t, ndim=3] mask,
                                             int level)
    cdef int fill_mask_selector(self, np.float64_t left_edge[3],
                                np.float64_t right_edge[3],
                                np.float64_t **dds, int dim[3],
                                np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask,
                                np.ndarray[np.uint8_t, ndim=3] mask,
                                int level)
    cdef void visit_grid_cells(self, GridVisitorData *data,
                    grid_visitor_function *func, np.uint8_t *cached_mask = ?)

    # compute periodic distance (if periodicity set)
    # assuming 0->domain_width[d] coordinates
    cdef np.float64_t periodic_difference(
        self, np.float64_t x1, np.float64_t x2, int d) nogil

cdef class AlwaysSelector(SelectorObject):
    pass

cdef class OctreeSubsetSelector(SelectorObject):
    cdef public SelectorObject base_selector
    cdef public np.int64_t domain_id

cdef class BooleanSelector(SelectorObject):
    cdef public SelectorObject sel1
    cdef public SelectorObject sel2

cdef inline np.float64_t _periodic_dist(np.float64_t x1, np.float64_t x2,
                                        np.float64_t dw, bint periodic) nogil:
    cdef np.float64_t rel = x1 - x2
    if not periodic: return rel
    if rel > dw * 0.5:
        rel -= dw
    elif rel < -dw * 0.5:
        rel += dw
    return rel



cdef class SliceSelector(SelectorObject):
    cdef public int axis
    cdef public np.float64_t coord
    cdef public int ax, ay
    cdef public int reduced_dimensionality

    def __init__(self, dobj):
        self.axis = dobj.axis
        self.coord = _ensure_code(dobj.coord)
        # If we have a reduced dimensionality dataset, we want to avoid any
        # checks against it in the axes that are beyond its dimensionality.
        # This means that if we have a 2D dataset, *all* slices along z will
        # select all the zones.
        if self.axis >= dobj.ds.dimensionality:
            self.reduced_dimensionality = 1
        else:
            self.reduced_dimensionality = 0

        self.ax = (self.axis+1) % 3
        self.ay = (self.axis+2) % 3

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    def fill_mask_regular_grid(self, gobj):
        cdef np.ndarray[np.uint8_t, ndim=3] mask
        cdef np.ndarray[np.uint8_t, ndim=3, cast=True] child_mask
        cdef int i, j, k
        cdef int total = 0
        cdef int this_level = 0
        cdef int ind[3][2]
        cdef np.uint64_t icoord
        cdef np.int32_t level = gobj.Level
        _ensure_code(gobj.LeftEdge)
        _ensure_code(gobj.dds)

        if level < self.min_level or level > self.max_level:
            return None
        else:
            child_mask = gobj.child_mask
            mask = np.zeros(gobj.ActiveDimensions, dtype=np.uint8)
            if level == self.max_level:
                this_level = 1
            for i in range(3):
                if i == self.axis:
                    icoord = <np.uint64_t>(
                        (self.coord - gobj.LeftEdge.d[i])/gobj.dds[i])
                    # clip coordinate to avoid seg fault below if we're
                    # exactly at a grid boundary
                    ind[i][0] = iclip(
                        icoord, 0, gobj.ActiveDimensions[i]-1)
                    ind[i][1] = ind[i][0] + 1
                else:
                    ind[i][0] = 0
                    ind[i][1] = gobj.ActiveDimensions[i]
            with nogil:
                for i in range(ind[0][0], ind[0][1]):
                    for j in range(ind[1][0], ind[1][1]):
                        for k in range(ind[2][0], ind[2][1]):
                            if this_level == 1 or child_mask[i, j, k]:
                                mask[i, j, k] = 1
                                total += 1
            if total == 0: return None
            return mask.astype("bool")

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil:
        if self.reduced_dimensionality == 1:
            return 1
        if pos[self.axis] + 0.5*dds[self.axis] > self.coord \
           and pos[self.axis] - 0.5*dds[self.axis] - grid_eps <= self.coord:
            return 1
        return 0

    cdef int select_point(self, np.float64_t pos[3]) noexcept nogil:
        # two 0-volume constructs don't intersect
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil:
        if self.reduced_dimensionality == 1:
            return 1
        cdef np.float64_t dist = self.periodic_difference(
            pos[self.axis], self.coord, self.axis)
        if dist*dist < radius*radius:
            return 1
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        if self.reduced_dimensionality == 1:
            return 1
        if left_edge[self.axis] - grid_eps <= self.coord < right_edge[self.axis]:
            return 1
        return 0

    @cython.boundscheck(False)
    @cython.wraparound(False)
    @cython.cdivision(True)
    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        if self.reduced_dimensionality == 1:
            return 2
        if left_edge[self.axis] - grid_eps <= self.coord < right_edge[self.axis]:
            return 2 # a box with non-zero volume can't be inside a plane
        return 0

    def _hash_vals(self):
        return (("axis", self.axis),
                ("coord", self.coord))

    def _get_state_attnames(self):
        return ("axis", "coord", "ax", "ay", "reduced_dimensionality")

slice_selector = SliceSelector



import yt

# Load the dataset.
ds = yt.load("Enzo_64/DD0043/data0043")
sc = yt.create_scene(ds, ("gas", "density"))

# You may need to adjust the alpha values to get a rendering with good contrast
# For annotate_domain, the fourth color value is alpha.

# Draw the domain boundary
sc.annotate_domain(ds, color=[1, 1, 1, 0.01])
sc.save(f"{ds}_vr_domain.png", sigma_clip=4)

# Draw the grid boundaries
sc.annotate_grids(ds, alpha=0.01)
sc.save(f"{ds}_vr_grids.png", sigma_clip=4)

# Draw a coordinate axes triad
sc.annotate_axes(alpha=0.01)
sc.save(f"{ds}_vr_coords.png", sigma_clip=4)



import numpy as np

from yt.frontends.sph.io import IOHandlerSPH
from yt.utilities.on_demand_imports import _h5py as h5py


class IOHandlerSwift(IOHandlerSPH):
    _dataset_type = "swift"

    def __init__(self, ds, *args, **kwargs):
        super().__init__(ds, *args, **kwargs)

    def _read_fluid_selection(self, chunks, selector, fields, size):
        raise NotImplementedError

    # NOTE: we refer to sub_files in the next sections, these sub_files may
    # actually be full data_files.
    # In the event data_files are too big, yt breaks them up into sub_files and
    # we sort of treat them as files in the chunking system
    def _read_particle_coords(self, chunks, ptf):
        # This will read chunks and yield the results.
        # yt has the concept of sub_files, i.e, we break up big files into
        # virtual sub_files to deal with the chunking system
        for sub_file in self._sorted_chunk_iterator(chunks):
            si, ei = sub_file.start, sub_file.end
            f = h5py.File(sub_file.filename, mode="r")
            # This double-reads
            for ptype in sorted(ptf):
                if sub_file.total_particles[ptype] == 0:
                    continue
                pos = f[f"/{ptype}/Coordinates"][si:ei, :]
                pos = pos.astype("float64", copy=False)
                if ptype == self.ds._sph_ptypes[0]:
                    hsml = self._get_smoothing_length(sub_file)
                else:
                    hsml = 0.0
                yield ptype, (pos[:, 0], pos[:, 1], pos[:, 2]), hsml
            f.close()

    def _yield_coordinates(self, sub_file, needed_ptype=None):
        si, ei = sub_file.start, sub_file.end
        f = h5py.File(sub_file.filename, mode="r")
        pcount = f["/Header"].attrs["NumPart_ThisFile"][:].astype("int")
        np.clip(pcount - si, 0, ei - si, out=pcount)
        pcount = pcount.sum()
        for key in f.keys():
            if (
                not key.startswith("PartType")
                or "Coordinates" not in f[key]
                or needed_ptype
                and key != needed_ptype
            ):
                continue
            pos = f[key]["Coordinates"][si:ei, ...]
            pos = pos.astype("float64", copy=False)
            yield key, pos
        f.close()

    def _get_smoothing_length(self, sub_file, pdtype=None, pshape=None):
        # We do not need the pdtype and the pshape, but some frontends do so we
        # accept them and then just ignore them
        ptype = self.ds._sph_ptypes[0]
        ind = int(ptype[-1])
        si, ei = sub_file.start, sub_file.end
        with h5py.File(sub_file.filename, mode="r") as f:
            pcount = f["/Header"].attrs["NumPart_ThisFile"][ind].astype("int")
            pcount = np.clip(pcount - si, 0, ei - si)
            # we upscale to float64
            hsml = f[ptype]["SmoothingLength"][si:ei, ...]
            hsml = hsml.astype("float64", copy=False)
            return hsml

    def _read_particle_data_file(self, sub_file, ptf, selector=None):
        # note: this frontend uses the variable name and terminology sub_file.
        # other frontends use data_file with the understanding that it may
        # actually be a sub_file, hence the super()._read_datafile is called
        # ._read_datafile instead of ._read_subfile
        return_data = {}

        si, ei = sub_file.start, sub_file.end
        f = h5py.File(sub_file.filename, mode="r")
        for ptype, field_list in sorted(ptf.items()):
            if sub_file.total_particles[ptype] == 0:
                continue
            g = f[f"/{ptype}"]
            # this should load as float64
            coords = g["Coordinates"][si:ei]
            if ptype == "PartType0":
                hsmls = self._get_smoothing_length(sub_file)
            else:
                hsmls = 0.0

            if selector:
                mask = selector.select_points(
                    coords[:, 0], coords[:, 1], coords[:, 2], hsmls
                )
            del coords
            if selector and mask is None:
                continue
            for field in field_list:
                if field in ("Mass", "Masses"):
                    data = g[self.ds._particle_mass_name][si:ei]
                else:
                    data = g[field][si:ei]

                if selector:
                    data = data[mask, ...]

                data.astype("float64", copy=False)

                return_data[(ptype, field)] = data
        f.close()

        return return_data

    def _count_particles(self, data_file):
        si, ei = data_file.start, data_file.end
        f = h5py.File(data_file.filename, mode="r")
        pcount = f["/Header"].attrs["NumPart_ThisFile"][:].astype("int")
        f.close()
        # if this data_file was a sub_file, then we just extract the region
        # defined by the subfile
        if None not in (si, ei):
            np.clip(pcount - si, 0, ei - si, out=pcount)
        npart = {f"PartType{i}": v for i, v in enumerate(pcount)}
        return npart

    def _identify_fields(self, data_file):
        f = h5py.File(data_file.filename, mode="r")
        fields = []
        cname = self.ds._particle_coordinates_name  # Coordinates
        mname = self.ds._particle_mass_name  # Coordinates

        for key in f.keys():
            if not key.startswith("PartType"):
                continue

            g = f[key]
            if cname not in g:
                continue

            ptype = str(key)
            for k in g.keys():
                kk = k
                if str(kk) == mname:
                    fields.append((ptype, "Mass"))
                    continue
                if not hasattr(g[kk], "shape"):
                    continue
                if len(g[kk].shape) > 1:
                    self._vector_fields[kk] = g[kk].shape[1]
                fields.append((ptype, str(kk)))

        f.close()
        return fields, {}



from yt.utilities.exceptions import YTModuleRemoved

raise YTModuleRemoved(
    "halo_analysis",
    "https://github.com/yt-project/yt_astro_analysis",
    "https://yt-astro-analysis.readthedocs.io/",
)



import numpy as np


def SIEVE_PRIMES(x):
    return x and x[:1] + SIEVE_PRIMES([n for n in x if n % x[0]])


def decompose_to_primes(max_prime):
    """Decompose number into the primes"""
    for prime in SIEVE_PRIMES(list(range(2, max_prime))):
        if prime * prime > max_prime:
            break
        while max_prime % prime == 0:
            yield prime
            max_prime //= prime
    if max_prime > 1:
        yield max_prime


def decompose_array(shape, psize, bbox):
    """Calculate list of product(psize) subarrays of arr, along with their
    left and right edges
    """
    return split_array(bbox[:, 0], bbox[:, 1], shape, psize)


def evaluate_domain_decomposition(n_d, pieces, ldom):
    """Evaluate longest to shortest edge ratio
    BEWARE: lot's of magic here"""
    eff_dim = (n_d > 1).sum()
    exp = float(eff_dim - 1) / float(eff_dim)
    ideal_bsize = eff_dim * pieces ** (1.0 / eff_dim) * np.prod(n_d) ** exp
    mask = np.where(n_d > 1)
    nd_arr = np.array(n_d, dtype=np.float64)[mask]
    bsize = int(np.sum(ldom[mask] / nd_arr * np.prod(nd_arr)))
    load_balance = float(np.prod(n_d)) / (
        float(pieces) * np.prod((n_d - 1) // ldom + 1)
    )

    # 0.25 is magic number
    quality = load_balance / (1 + 0.25 * (bsize / ideal_bsize - 1.0))
    # \todo add a factor that estimates lower cost when x-direction is
    # not chopped too much
    # \deprecated estimate these magic numbers
    quality *= 1.0 - (0.001 * ldom[0] + 0.0001 * ldom[1]) / pieces
    if np.any(ldom > n_d):
        quality = 0

    return quality


def factorize_number(pieces):
    """Return array consisting of prime, its power and number of different
    decompositions in three dimensions for this prime
    """
    factors = list(decompose_to_primes(pieces))
    temp = np.bincount(factors)
    return np.array(
        [
            (prime, temp[prime], (temp[prime] + 1) * (temp[prime] + 2) // 2)
            for prime in np.unique(factors)
        ]
    ).astype(np.int64)


def get_psize(n_d, pieces):
    """Calculate the best division of array into px*py*pz subarrays.
    The goal is to minimize the ratio of longest to shortest edge
    to minimize the amount of inter-process communication.
    """
    fac = factorize_number(pieces)
    nfactors = len(fac[:, 2])
    best = 0.0
    p_size = np.ones(3, dtype=np.int64)
    if pieces == 1:
        return p_size

    while np.all(fac[:, 2] > 0):
        ldom = np.ones(3, dtype=np.int64)
        for nfac in range(nfactors):
            i = int(np.sqrt(0.25 + 2 * (fac[nfac, 2] - 1)) - 0.5)
            k = fac[nfac, 2] - (1 + i * (i + 1) // 2)
            i = fac[nfac, 1] - i
            j = fac[nfac, 1] - (i + k)
            ldom *= fac[nfac, 0] ** np.array([i, j, k])

        quality = evaluate_domain_decomposition(n_d, pieces, ldom)
        if quality > best:
            best = quality
            p_size = ldom
        # search for next unique combination
        for j in range(nfactors):
            if fac[j, 2] > 1:
                fac[j, 2] -= 1
                break
            else:
                if j < nfactors - 1:
                    fac[j, 2] = int((fac[j, 1] + 1) * (fac[j, 1] + 2) / 2)
                else:
                    fac[:, 2] = 0  # no more combinations to try

    return p_size


def split_array(gle, gre, shape, psize):
    """Split array into px*py*pz subarrays."""
    n_d = np.array(shape, dtype=np.int64)
    dds = (gre - gle) / shape
    left_edges = []
    right_edges = []
    shapes = []
    slices = []
    for i in range(psize[0]):
        for j in range(psize[1]):
            for k in range(psize[2]):
                piece = np.array((i, j, k), dtype=np.int64)
                lei = n_d * piece // psize
                rei = n_d * (piece + np.ones(3, dtype=np.int64)) // psize
                lle = gle + lei * dds
                lre = gle + rei * dds
                left_edges.append(lle)
                right_edges.append(lre)
                shapes.append(rei - lei)
                slices.append(np.s_[lei[0] : rei[0], lei[1] : rei[1], lei[2] : rei[2]])

    return left_edges, right_edges, shapes, slices



import collections

templates = dict(
    autoclass=r"""
%(name)s
%(header)s

.. autoclass:: %(name)s
   :members:
   :inherited-members:
   :undoc-members:

""",
    autofunction=r"""
%(name)s
%(header)s

.. autofunction:: %(name)s

""",
    index_file=r"""
%(title)s
%(header)s

.. autosummary::
   :toctree: generated/%(dn)s

""",
)

file_names = dict(
    ft=("Field Types", "source/api/field_types/%s.rst"),
    pt=("Plot Types", "source/api/plot_types/%s.rst"),
    cl=("Callback List", "source/api/callback_list/%s.rst"),
    ee=("Extension Types", "source/api/extension_types/%s.rst"),
    dd=("Derived Datatypes", "source/api/derived_datatypes/%s.rst"),
    mt=("Miscellaneous Types", "source/api/misc_types/%s.rst"),
    fl=("Function List", "source/api/function_list/%s.rst"),
    ds=("Data Sources", "source/api/data_sources/%s.rst"),
    dq=("Derived Quantities", "source/api/derived_quantities/%s.rst"),
)

to_include = collections.defaultdict(list)

for line in open("auto_generated.txt"):
    ftype, name, file_name = (s.strip() for s in line.split("::"))
    cn = name.split(".")[-1]
    if cn[0] == "_":
        cn = cn[1:]  # For leading _
    fn = file_names[file_name][1] % cn
    # if not os.path.exists(os.path.dirname(fn)):
    #    os.mkdir(os.path.dirname(fn))
    header = "-" * len(name)
    dd = dict(header=header, name=name)
    # open(fn, "w").write(templates[ftype] % dd)
    to_include[file_name].append(name)

for key, val in file_names.items():
    title, file = val
    fn = file.rsplit("/", 1)[0] + ".rst"
    print(fn)
    f = open(fn, "w")
    dn = fn.split("/")[-1][:-4]
    dd = dict(header="=" * len(title), title=title, dn=dn)
    f.write(templates["index_file"] % dd)
    for obj in sorted(to_include[key]):
        f.write(f"   {obj}\n")



"""
A proxy object for field descriptors, usually living as ds.fields.
"""

import inspect
import textwrap
import weakref
from functools import cached_property

from yt.fields.derived_field import DerivedField


def _fill_values(values):
    value = (
        '<div class="rendered_html jp-RenderedHTMLCommon">'
        + "<table><thead><tr><th>Name</th><th>Type</th>"
        + "<th>Value</th></tr></thead><tr><td>"
        + "</td></tr><tr><td>".join(
            "{}</td><td>{}</td><td>{}".format(
                v, type(values[v]).__name__, str(values[v])
            )
            for v in sorted(values)
        )
        + "</td></tr></table></div>"
    )
    return value


class FieldTypeContainer:
    def __init__(self, ds):
        self.ds = weakref.proxy(ds)

    def __getattr__(self, attr):
        ds = self.__getattribute__("ds")
        fnc = FieldNameContainer(ds, attr)
        if len(dir(fnc)) == 0:
            return self.__getattribute__(attr)
        return fnc

    @cached_property
    def field_types(self):
        return {t for t, n in self.ds.field_info}

    def __dir__(self):
        return list(self.field_types)

    def __iter__(self):
        for ft in self.field_types:
            fnc = FieldNameContainer(self.ds, ft)
            if len(dir(fnc)) == 0:
                yield self.__getattribute__(ft)
            else:
                yield fnc

    def __contains__(self, obj):
        ob = None
        if isinstance(obj, FieldNameContainer):
            ob = obj.field_type
        elif isinstance(obj, str):
            ob = obj

        return ob in self.field_types

    def _ipython_display_(self):
        import ipywidgets
        from IPython.display import display

        fnames = []
        children = []
        for ftype in sorted(self.field_types):
            fnc = getattr(self, ftype)
            children.append(ipywidgets.Output())
            with children[-1]:
                display(fnc)
            fnames.append(ftype)
        tabs = ipywidgets.Tab(children=children)
        for i, n in enumerate(fnames):
            tabs.set_title(i, n)
        display(tabs)


class FieldNameContainer:
    def __init__(self, ds, field_type):
        self.ds = ds
        self.field_type = field_type

    def __getattr__(self, attr):
        ft = self.__getattribute__("field_type")
        ds = self.__getattribute__("ds")
        if (ft, attr) not in ds.field_info:
            return self.__getattribute__(attr)
        return ds.field_info[ft, attr]

    def __dir__(self):
        return [n for t, n in self.ds.field_info if t == self.field_type]

    def __iter__(self):
        for t, n in self.ds.field_info:
            if t == self.field_type:
                yield self.ds.field_info[t, n]

    def __contains__(self, obj):
        if isinstance(obj, DerivedField):
            if self.field_type == obj.name[0] and obj.name in self.ds.field_info:
                # e.g. from a completely different dataset
                if self.ds.field_info[obj.name] is not obj:
                    return False
                return True
        elif isinstance(obj, tuple):
            if self.field_type == obj[0] and obj in self.ds.field_info:
                return True
        elif isinstance(obj, str):
            if (self.field_type, obj) in self.ds.field_info:
                return True
        return False

    def _ipython_display_(self):
        import ipywidgets
        from IPython.display import Markdown, display

        names = dir(self)
        names.sort()

        def change_field(_ftype, _box, _var_window):
            def _change_field(event):
                fobj = getattr(_ftype, event["new"])
                _box.clear_output()
                with _box:
                    display(
                        Markdown(
                            data="```python\n"
                            + textwrap.dedent(fobj.get_source())
                            + "\n```"
                        )
                    )
                values = inspect.getclosurevars(fobj._function).nonlocals
                _var_window.value = _fill_values(values)

            return _change_field

        flist = ipywidgets.Select(options=names, layout=ipywidgets.Layout(height="95%"))
        source = ipywidgets.Output(layout=ipywidgets.Layout(width="100%", height="9em"))
        var_window = ipywidgets.HTML(value="Empty")
        var_box = ipywidgets.Box(
            layout=ipywidgets.Layout(width="100%", height="100%", overflow_y="scroll")
        )
        var_box.children = [var_window]
        ftype_tabs = ipywidgets.Tab(
            children=[source, var_box],
            layout=ipywidgets.Layout(flex="2 1 auto", width="auto", height="95%"),
        )
        ftype_tabs.set_title(0, "Source")
        ftype_tabs.set_title(1, "Variables")
        flist.observe(change_field(self, source, var_window), "value")
        display(
            ipywidgets.HBox(
                [flist, ftype_tabs], layout=ipywidgets.Layout(height="14em")
            )
        )



import abc
import os
import weakref
from typing import Optional, Tuple

import numpy as np

from yt._maintenance.deprecation import issue_deprecation_warning
from yt.config import ytcfg
from yt.units._numpy_wrapper_functions import uconcatenate
from yt.units.yt_array import YTArray
from yt.utilities.exceptions import YTFieldNotFound
from yt.utilities.io_handler import io_registry
from yt.utilities.logger import ytLogger as mylog
from yt.utilities.on_demand_imports import _h5py as h5py
from yt.utilities.parallel_tools.parallel_analysis_interface import (
    ParallelAnalysisInterface,
    parallel_root_only,
)


class Index(ParallelAnalysisInterface, abc.ABC):
    """The base index class"""

    _unsupported_objects: Tuple[str, ...] = ()
    _index_properties: Tuple[str, ...] = ()

    def __init__(self, ds, dataset_type):
        ParallelAnalysisInterface.__init__(self)
        self.dataset = weakref.proxy(ds)
        self.ds = self.dataset

        self._initialize_state_variables()

        mylog.debug("Initializing data storage.")
        self._initialize_data_storage()

        mylog.debug("Setting up domain geometry.")
        self._setup_geometry()

        mylog.debug("Initializing data grid data IO")
        self._setup_data_io()

        # Note that this falls under the "geometry" object since it's
        # potentially quite expensive, and should be done with the indexing.
        mylog.debug("Detecting fields.")
        self._detect_output_fields()

    @abc.abstractmethod
    def _detect_output_fields(self):
        pass

    def _icoords_to_fcoords(
        self,
        icoords: np.ndarray,
        ires: np.ndarray,
        axes: Optional[Tuple[int, ...]] = None,
    ) -> Tuple[np.ndarray, np.ndarray]:
        # What's the use of raising NotImplementedError for this, when it's an
        # abstract base class?  Well, only *some* of the subclasses have it --
        # and for those that *don't*, we should not be calling it -- and since
        # it's a semi-private method, it shouldn't be called outside of yt
        # machinery.  So we shouldn't ever get here!
        raise NotImplementedError

    def _initialize_state_variables(self):
        self._parallel_locking = False
        self._data_file = None
        self._data_mode = None
        self.num_grids = None

    def _initialize_data_storage(self):
        if not ytcfg.get("yt", "serialize"):
            return
        fn = self.ds.storage_filename
        if fn is None:
            if os.path.isfile(
                os.path.join(self.directory, f"{self.ds.unique_identifier}.yt")
            ):
                fn = os.path.join(self.directory, f"{self.ds.unique_identifier}.yt")
            else:
                fn = os.path.join(self.directory, f"{self.dataset.basename}.yt")
        dir_to_check = os.path.dirname(fn)
        if dir_to_check == "":
            dir_to_check = "."
        # We have four options:
        #    Writeable, does not exist      : create, open as append
        #    Writeable, does exist          : open as append
        #    Not writeable, does not exist  : do not attempt to open
        #    Not writeable, does exist      : open as read-only
        exists = os.path.isfile(fn)
        if not exists:
            writeable = os.access(dir_to_check, os.W_OK)
        else:
            writeable = os.access(fn, os.W_OK)
        writeable = writeable and not ytcfg.get("yt", "only_deserialize")
        # We now have our conditional stuff
        self.comm.barrier()
        if not writeable and not exists:
            return
        if writeable:
            try:
                if not exists:
                    self.__create_data_file(fn)
                self._data_mode = "a"
            except OSError:
                self._data_mode = None
                return
        else:
            self._data_mode = "r"

        self.__data_filename = fn
        self._data_file = h5py.File(fn, mode=self._data_mode)

    def __create_data_file(self, fn):
        # Note that this used to be parallel_root_only; it no longer is,
        # because we have better logic to decide who owns the file.
        f = h5py.File(fn, mode="a")
        f.close()

    def _setup_data_io(self):
        if getattr(self, "io", None) is not None:
            return
        self.io = io_registry[self.dataset_type](self.dataset)

    @parallel_root_only
    def save_data(
        self, array, node, name, set_attr=None, force=False, passthrough=False
    ):
        """
        Arbitrary numpy data will be saved to the region in the datafile
        described by *node* and *name*.  If data file does not exist, it throws
        no error and simply does not save.
        """

        if self._data_mode != "a":
            return
        try:
            node_loc = self._data_file[node]
            if name in node_loc and force:
                mylog.info("Overwriting node %s/%s", node, name)
                del self._data_file[node][name]
            elif name in node_loc and passthrough:
                return
        except Exception:
            pass
        myGroup = self._data_file["/"]
        for q in node.split("/"):
            if q:
                myGroup = myGroup.require_group(q)
        arr = myGroup.create_dataset(name, data=array)
        if set_attr is not None:
            for i, j in set_attr.items():
                arr.attrs[i] = j
        self._data_file.flush()

    def _reload_data_file(self, *args, **kwargs):
        if self._data_file is None:
            return
        self._data_file.close()
        del self._data_file
        self._data_file = h5py.File(self.__data_filename, mode=self._data_mode)

    def get_data(self, node, name):
        """
        Return the dataset with a given *name* located at *node* in the
        datafile.
        """
        if self._data_file is None:
            return None
        if node[0] != "/":
            node = f"/{node}"

        myGroup = self._data_file["/"]
        for group in node.split("/"):
            if group:
                if group not in myGroup:
                    return None
                myGroup = myGroup[group]
        if name not in myGroup:
            return None

        full_name = f"{node}/{name}"
        try:
            return self._data_file[full_name][:]
        except TypeError:
            return self._data_file[full_name]

    def _get_particle_type_counts(self):
        # this is implemented by subclasses
        raise NotImplementedError

    def _close_data_file(self):
        if self._data_file:
            self._data_file.close()
            del self._data_file
            self._data_file = None

    def _split_fields(self, fields):
        # This will split fields into either generated or read fields
        fields_to_read, fields_to_generate = [], []
        for ftype, fname in fields:
            if fname in self.field_list or (ftype, fname) in self.field_list:
                fields_to_read.append((ftype, fname))
            elif (
                fname in self.ds.derived_field_list
                or (ftype, fname) in self.ds.derived_field_list
            ):
                fields_to_generate.append((ftype, fname))
            else:
                raise YTFieldNotFound((ftype, fname), self.ds)
        return fields_to_read, fields_to_generate

    def _read_particle_fields(self, fields, dobj, chunk=None):
        if len(fields) == 0:
            return {}, []
        fields_to_read, fields_to_generate = self._split_fields(fields)
        if len(fields_to_read) == 0:
            return {}, fields_to_generate
        selector = dobj.selector
        if chunk is None:
            self._identify_base_chunk(dobj)
        chunks = self._chunk_io(dobj, cache=False)
        fields_to_return = self.io._read_particle_selection(
            chunks, selector, fields_to_read
        )
        return fields_to_return, fields_to_generate

    def _read_fluid_fields(self, fields, dobj, chunk=None):
        if len(fields) == 0:
            return {}, []
        fields_to_read, fields_to_generate = self._split_fields(fields)
        if len(fields_to_read) == 0:
            return {}, fields_to_generate
        selector = dobj.selector
        if chunk is None:
            self._identify_base_chunk(dobj)
            chunk_size = dobj.size
        else:
            chunk_size = chunk.data_size
        fields_to_return = self.io._read_fluid_selection(
            self._chunk_io(dobj), selector, fields_to_read, chunk_size
        )
        return fields_to_return, fields_to_generate

    def _chunk(self, dobj, chunking_style, ngz=0, **kwargs):
        # A chunk is either None or (grids, size)
        if dobj._current_chunk is None:
            self._identify_base_chunk(dobj)
        if ngz != 0 and chunking_style != "spatial":
            raise NotImplementedError
        if chunking_style == "all":
            return self._chunk_all(dobj, **kwargs)
        elif chunking_style == "spatial":
            return self._chunk_spatial(dobj, ngz, **kwargs)
        elif chunking_style == "io":
            return self._chunk_io(dobj, **kwargs)
        else:
            raise NotImplementedError


def cacheable_property(func):
    # not quite equivalent to functools.cached_property
    # this decorator allows cached to be disabled via a self._cache flag attribute
    n = f"_{func.__name__}"

    @property
    def cacheable_func(self):
        if self._cache and getattr(self, n, None) is not None:
            return getattr(self, n)
        if self.data_size is None:
            tr = self._accumulate_values(n[1:])
        else:
            tr = func(self)
        if self._cache:
            setattr(self, n, tr)
        return tr

    return cacheable_func


class YTDataChunk:
    def __init__(
        self,
        dobj,
        chunk_type,
        objs,
        data_size=None,
        field_type=None,
        cache=False,
        fast_index=None,
    ):
        self.dobj = dobj
        self.chunk_type = chunk_type
        self.objs = objs
        self.data_size = data_size
        self._field_type = field_type
        self._cache = cache
        self._fast_index = fast_index

    def _accumulate_values(self, method):
        # We call this generically.  It's somewhat slower, since we're doing
        # costly getattr functions, but this allows us to generalize.
        mname = f"select_{method}"
        arrs = []
        for obj in self._fast_index or self.objs:
            f = getattr(obj, mname)
            arrs.append(f(self.dobj))
        if method == "dtcoords":
            arrs = [arr[0] for arr in arrs]
        elif method == "tcoords":
            arrs = [arr[1] for arr in arrs]
        arrs = uconcatenate(arrs)
        self.data_size = arrs.shape[0]
        return arrs

    @cacheable_property
    def fcoords(self):
        if self._fast_index is not None:
            ci = self._fast_index.select_fcoords(self.dobj.selector, self.data_size)
            ci = YTArray(ci, units="code_length", registry=self.dobj.ds.unit_registry)
            return ci
        ci = np.empty((self.data_size, 3), dtype="float64")
        ci = YTArray(ci, units="code_length", registry=self.dobj.ds.unit_registry)
        if self.data_size == 0:
            return ci
        ind = 0
        for obj in self._fast_index or self.objs:
            c = obj.select_fcoords(self.dobj)
            if c.shape[0] == 0:
                continue
            ci.d[ind : ind + c.shape[0], :] = c
            ind += c.shape[0]
        return ci

    @cacheable_property
    def icoords(self):
        if self._fast_index is not None:
            ci = self._fast_index.select_icoords(self.dobj.selector, self.data_size)
            return ci
        ci = np.empty((self.data_size, 3), dtype="int64")
        if self.data_size == 0:
            return ci
        ind = 0
        for obj in self._fast_index or self.objs:
            c = obj.select_icoords(self.dobj)
            if c.shape[0] == 0:
                continue
            ci[ind : ind + c.shape[0], :] = c
            ind += c.shape[0]
        return ci

    @cacheable_property
    def fwidth(self):
        if self._fast_index is not None:
            ci = self._fast_index.select_fwidth(self.dobj.selector, self.data_size)
            ci = YTArray(ci, units="code_length", registry=self.dobj.ds.unit_registry)
            return ci
        ci = np.empty((self.data_size, 3), dtype="float64")
        ci = YTArray(ci, units="code_length", registry=self.dobj.ds.unit_registry)
        if self.data_size == 0:
            return ci
        ind = 0
        for obj in self._fast_index or self.objs:
            c = obj.select_fwidth(self.dobj)
            if c.shape[0] == 0:
                continue
            ci.d[ind : ind + c.shape[0], :] = c

            ind += c.shape[0]
        return ci

    @cacheable_property
    def ires(self):
        if self._fast_index is not None:
            ci = self._fast_index.select_ires(self.dobj.selector, self.data_size)
            return ci
        ci = np.empty(self.data_size, dtype="int64")
        if self.data_size == 0:
            return ci
        ind = 0
        for obj in self._fast_index or self.objs:
            c = obj.select_ires(self.dobj)
            if c.shape == 0:
                continue
            ci[ind : ind + c.size] = c
            ind += c.size
        return ci

    @cacheable_property
    def tcoords(self):
        self.dtcoords
        return self._tcoords

    @cacheable_property
    def dtcoords(self):
        ct = np.empty(self.data_size, dtype="float64")
        cdt = np.empty(self.data_size, dtype="float64")
        self._tcoords = ct  # Se this for tcoords
        if self.data_size == 0:
            return cdt
        ind = 0
        for obj in self._fast_index or self.objs:
            gdt, gt = obj.select_tcoords(self.dobj)
            if gt.size == 0:
                continue
            ct[ind : ind + gt.size] = gt
            cdt[ind : ind + gdt.size] = gdt
            ind += gt.size
        return cdt

    @cacheable_property
    def fcoords_vertex(self):
        nodes_per_elem = self.dobj.index.meshes[0].connectivity_indices.shape[1]
        dim = self.dobj.ds.dimensionality
        ci = np.empty((self.data_size, nodes_per_elem, dim), dtype="float64")
        ci = YTArray(ci, units="code_length", registry=self.dobj.ds.unit_registry)
        if self.data_size == 0:
            return ci
        ind = 0
        for obj in self.objs:
            c = obj.select_fcoords_vertex(self.dobj)
            if c.shape[0] == 0:
                continue
            ci.d[ind : ind + c.shape[0], :, :] = c
            ind += c.shape[0]
        return ci


class ChunkDataCache:
    def __init__(self, base_iter, preload_fields, geometry_handler, max_length=256):
        # At some point, max_length should instead become a heuristic function,
        # potentially looking at estimated memory usage.  Note that this never
        # initializes the iterator; it assumes the iterator is already created,
        # and it calls next() on it.
        self.base_iter = base_iter.__iter__()
        self.queue = []
        self.max_length = max_length
        self.preload_fields = preload_fields
        self.geometry_handler = geometry_handler
        self.cache = {}

    def __iter__(self):
        return self

    def __next__(self):
        if len(self.queue) == 0:
            for _ in range(self.max_length):
                try:
                    self.queue.append(next(self.base_iter))
                except StopIteration:
                    break
            # If it's still zero ...
            if len(self.queue) == 0:
                raise StopIteration
            chunk = YTDataChunk(None, "cache", self.queue, cache=False)
            self.cache = (
                self.geometry_handler.io._read_chunk_data(chunk, self.preload_fields)
                or {}
            )
        g = self.queue.pop(0)
        g._initialize_cache(self.cache.pop(g.id, {}))
        return g


def is_curvilinear(geo):
    # tell geometry is curvilinear or not
    issue_deprecation_warning(
        "the is_curvilear() function is deprecated. "
        "Instead, compare the geometry object directly with yt.geometry.geometry_enum.Geometry "
        "enum members, as for instance:\n"
        "if is_curvilinear(geometry):\n    ...\n"
        "should be rewritten as:"
        "if geometry is Geometry.POLAR or geometry is Geometry.CYLINDRICAL or geometry is Geometry.SPHERICAL:\n    ...",
        since="4.2",
    )
    if geo in ["polar", "cylindrical", "spherical"]:
        return True
    else:
        return False



import pytest

from yt.testing import fake_amr_ds


def test_ambiguous_fails():
    ds = fake_amr_ds(particles=10)
    msg = "The requested field name '{}' is ambiguous"

    def _mock_field(field, data):
        return data["ones"]

    # create a pair of ambiguous field
    ds.add_field(("io", "mock_field"), _mock_field, sampling_type="particle")
    ds.add_field(("gas", "mock_field"), _mock_field, sampling_type="cell")

    # Test errors are raised for ambiguous fields
    with pytest.raises(ValueError, match=msg.format("mock_field")):
        ds.r["mock_field"]

    # check that explicit name tuples don't raise a warning
    ds.r["io", "mock_field"]
    ds.r["gas", "mock_field"]


def test_nameonly_field_with_all_aliases_candidates():
    # see https://github.com/yt-project/yt/issues/3839
    ds = fake_amr_ds(fields=["density"], units=["g/cm**3"])

    # here we rely on implementations details from fake_amr_ds,
    # so we verify that it provides the appropriate conditions
    # for the actual test.
    candidates = [f for f in ds.derived_field_list if f[1] == "density"]
    assert len(candidates) == 2
    fi = ds.field_info
    assert fi[candidates[0]].is_alias_to(fi[candidates[1]])

    # this is the actual test (check that no error or warning is raised)
    ds.all_data()["density"]



import numpy as np

from yt.frontends.gadget.api import GadgetHDF5Dataset
from yt.funcs import mylog
from yt.utilities.on_demand_imports import _h5py as h5py

from .fields import ArepoFieldInfo


class ArepoHDF5Dataset(GadgetHDF5Dataset):
    _field_info_class = ArepoFieldInfo

    def __init__(
        self,
        filename,
        dataset_type="arepo_hdf5",
        unit_base=None,
        smoothing_factor=2.0,
        index_order=None,
        index_filename=None,
        kernel_name=None,
        bounding_box=None,
        units_override=None,
        unit_system="cgs",
        default_species_fields=None,
    ):
        super().__init__(
            filename,
            dataset_type=dataset_type,
            unit_base=unit_base,
            index_order=index_order,
            index_filename=index_filename,
            kernel_name=kernel_name,
            bounding_box=bounding_box,
            units_override=units_override,
            unit_system=unit_system,
            default_species_fields=default_species_fields,
        )
        # The "smoothing_factor" is a user-configurable parameter which
        # is multiplied by the radius of the sphere with a volume equal
        # to that of the Voronoi cell to create smoothing lengths.
        self.smoothing_factor = smoothing_factor
        self.gamma = 5.0 / 3.0
        self.gamma_cr = self.parameters.get("GammaCR", 4.0 / 3.0)

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        need_groups = ["Header", "Config"]
        veto_groups = ["FOF", "Group", "Subhalo"]
        valid = True
        try:
            fh = h5py.File(filename, mode="r")
            valid = (
                all(ng in fh["/"] for ng in need_groups)
                and not any(vg in fh["/"] for vg in veto_groups)
                and (
                    "VORONOI" in fh["/Config"].attrs.keys()
                    or "AMR" in fh["/Config"].attrs.keys()
                )
                # Datasets with GFM_ fields present are AREPO
                or any(field.startswith("GFM_") for field in fh["/PartType0"])
            )
            fh.close()
        except Exception:
            valid = False
        return valid

    def _get_uvals(self):
        handle = h5py.File(self.parameter_filename, mode="r")
        uvals = {}
        missing = [True] * 3
        for i, unit in enumerate(
            ["UnitLength_in_cm", "UnitMass_in_g", "UnitVelocity_in_cm_per_s"]
        ):
            for grp in ["Header", "Parameters", "Units"]:
                if grp in handle and unit in handle[grp].attrs:
                    uvals[unit] = handle[grp].attrs[unit]
                    missing[i] = False
                    break
        if "UnitLength_in_cm" in uvals:
            # We assume this is comoving, because in the absence of comoving
            # integration the redshift will be zero.
            uvals["cmcm"] = 1.0 / uvals["UnitLength_in_cm"]
        handle.close()
        if all(missing):
            uvals = None
        return uvals

    def _set_code_unit_attributes(self):
        arepo_unit_base = self._get_uvals()
        # This rather convoluted logic is required to ensure that
        # units which are present in the Arepo dataset will be used
        # no matter what but that the user gets warned
        if arepo_unit_base is not None:
            if self._unit_base is None:
                self._unit_base = arepo_unit_base
            else:
                for unit in arepo_unit_base:
                    if unit == "cmcm":
                        continue
                    short_unit = unit.split("_")[0][4:].lower()
                    if short_unit in self._unit_base:
                        which_unit = short_unit
                        self._unit_base.pop(short_unit, None)
                    elif unit in self._unit_base:
                        which_unit = unit
                    else:
                        which_unit = None
                    if which_unit is not None:
                        msg = f"Overwriting '{which_unit}' in unit_base with what we found in the dataset."
                        mylog.warning(msg)
                    self._unit_base[unit] = arepo_unit_base[unit]
                if "cmcm" in arepo_unit_base:
                    self._unit_base["cmcm"] = arepo_unit_base["cmcm"]
        super()._set_code_unit_attributes()
        munit = np.sqrt(self.mass_unit / (self.time_unit**2 * self.length_unit)).to(
            "gauss"
        )
        if self.cosmological_simulation:
            self.magnetic_unit = self.quan(munit.value, f"{munit.units}/a**2")
        else:
            self.magnetic_unit = munit



"""
CF-radial-specific fields



"""
from typing import List

from yt.fields.field_info_container import FieldInfoContainer


class CFRadialFieldInfo(FieldInfoContainer):
    known_other_fields = ()  # fields are set dynamically
    known_particle_fields = ()
    units_to_ignore = ("dBz", "dBZ", "ratio")  # set as nondimensional if found
    field_units_ignored: List[str] = []  # fields for which units have been ignored

    # (find, replace) pairs for sanitizing:
    unit_subs = (("degrees", "degree"), ("meters", "m"), ("_per_", "/"))

    def setup_fluid_fields(self):
        # Here we dynamically add fields available in our netcdf file for to the
        # FieldInfoContainer with sanitized units.

        for field in self.field_list:  # field here is ('fluid_type', 'field') tuple
            units = self.ds.field_units.get(field, "")

            # sanitization of the units
            if units in self.units_to_ignore:
                self.field_units_ignored.append(field)
                units = ""

            for findstr, repstr in self.unit_subs:
                units = units.replace(findstr, repstr)

            self.add_output_field(field, "cell", units=units)
